import os
import re
import sys
import time
import datetime
from .MiscUtils import commands_get_status_output
try:
    long()
except Exception:
    long = int

from . import PLogger
from .LocalJobSpec import LocalJobSpec
from .LocalJobsetSpec import LocalJobsetSpec

class PdbProxy:

    # constructor
    def __init__(self,verbose=False):
        # database engine
        self.engine = 'sqlite3'
        # version of database schema
        self.version = '0_0_1'
        # database file name
        self.filename = 'pandajob.db'
        # database dir
        self.database_dir = os.path.expanduser(os.environ['PANDA_CONFIG_ROOT'])
        # full path of database file
        self.database = '%s/%s' % (self.database_dir,self.filename)
        # table name
        self.tablename = 'jobtable_%s' % self.version
        # verbose
        self.verbose = verbose
        # connection
        self.con = None
        # logger
        self.log = PLogger.getPandaLogger()


    # set verbose
    def setVerbose(self,verbose):
        # verbose
        self.verbose = verbose


    # execute SQL
    def execute(self,sql,var={}):
        # logger
        tmpLog = PLogger.getPandaLogger()
        # expand variables
        for tmpKey in var:
            tmpVal = var[tmpKey]
            sql = sql.replqce(tmpKey,str(tmpVal))
        # construct command
        com = '%s %s "%s"' % (self.engine,self.database,sql)
        if self.verbose:
            tmpLog.debug("DB Req : " + com)
        # execute
        nTry = 5
        status =0
        for iTry in range(nTry):
            if self.verbose:
                tmpLog.debug("   Try : %s/%s" % (iTry,nTry))
            status,output = commands_get_status_output(com)
            status %= 255
            if status == 0:
                break
            if iTry+1 < nTry:
                time.sleep(2)
        # return
        if status != 0:
            tmpLog.error(status)
            tmpLog.error(output)
            return False,output
        else:
            if self.verbose:
                tmpLog.debug("   Ret : " + output)
            outList = output.split('\n')
            # remove ''
            try:
                outList.remove('')
            except Exception:
                pass
            # remove junk messages
            ngStrings = ['Loading resources from']
            for tmpStr in tuple(outList):
                # look for NG strings
                flagNG = False
                for ngStr in ngStrings:
                    match = re.search(ngStr,tmpStr,re.I)
                    if match is not None:
                        flagNG = True
                        break
                # remove
                if flagNG:
                    try:
                        outList.remove(tmpStr)
                    except Exception:
                        pass
            return True,outList

    # execute SQL
    def execute_direct(self, sql, var=None, fetch=False):
        if self.con is None:
            import sqlite3
            self.con = sqlite3.connect(self.database, check_same_thread=False)
        if self.verbose:
            self.log.debug("DB Req : {0} var={1}".format(sql, str(var)))
        cur = self.con.cursor()
        try:
            if var is None:
                var = {}
            cur.execute(sql, var)
            retVal = True
        except Exception:
            retVal = False
            if not self.verbose:
                self.log.error("DB Req : {0} var={1}".format(sql, str(var)))
            err_type, err_value = sys.exc_info()[:2]
            err_str = "{0} {1}".format(err_type.__name__, err_value)
            self.log.error(err_str)
        if self.verbose:
            self.log.debug(retVal)
        outList = []
        if retVal:
            if fetch:
                outList = cur.fetchall()
                if self.verbose:
                    for item in outList:
                        self.log.debug("   Ret : " + str(item))
        self.con.commit()
        return retVal, outList


    # remove old database
    def deleteDatabase(self):
        commands_get_status_output('rm -f %s' % self.database)


    # initialize database
    def initialize(self):
        # import sqlite3
        # check if sqlite3 is available
        com = 'which %s' % self.engine
        status,output = commands_get_status_output(com)
        if status != 0:
            errstr  = "\n\n"
            errstr += "ERROR : %s is not available in PATH\n\n" % self.engine
            errstr += "There are some possible solutions\n"
            errstr += " * run this application under Athena runtime with Release 14 or higher. e.g.,\n"
            errstr += "   $ source setup.sh -tag=14.2.24,32,setup\n"
            errstr += "   $ source .../etc/panda/panda_setup.sh\n\n"
            errstr += " * set PATH and LD_LIBRARY_PATH to include %s. e.g., at CERN\n" % self.engine
            errstr += "   $ export PATH=/afs/cern.ch/sw/lcg/external/sqlite/3.4.0/slc3_ia32_gcc323/bin:$PATH\n"
            errstr += "   $ export LD_LIBRARY_PATH=/afs/cern.ch/sw/lcg/external/sqlite/3.4.0/slc3_ia32_gcc323/lib:$LD_LIBRARY_PATH\n"
            errstr += "   $ source .../etc/panda/panda_setup.sh\n\n"
            errstr += " * install %s from the standard SL4 repository. e.g.,\n" % self.engine
            errstr += "   $ yum install %s\n\n" % self.engine
            errstr += " * use SLC5\n"
            raise RuntimeError(errstr)
        # create dir for DB
        if not os.path.exists(self.database_dir):
            os.makedirs(self.database_dir)
        # the table already exist
        if self.checkTable():
            return
        # create table
        self.createTable()
        return


    # check table
    def checkTable(self):
        # get tables
        retS,retV = self.execute('.table')
        if not retS:
            raise RuntimeError("cannot get tables")
        # the table already exist or not
        if retV == []:
            return False
        if self.tablename not in retV[-1].split():
            return False
        # check schema
        self.checkSchema()
        return True


    # check schema
    def checkSchema(self,noAdd=False):
        # get colum names
        retS,retV = self.execute('PRAGMA table_info(%s)' % self.tablename)
        if not retS:
            raise RuntimeError("cannot get table_info")
        # parse
        columns = []
        for line in retV:
            items = line.split('|')
            if len(items) > 1:
                columns.append(items[1])
        # check
        for tmpC in LocalJobSpec.appended:
            tmpA = LocalJobSpec.appended[tmpC]
            if tmpC not in columns:
                if noAdd:
                    raise RuntimeError("%s not found in database schema" % tmpC)
                # add column
                retS,retV = self.execute("ALTER TABLE %s ADD COLUMN '%s' %s" % \
                                         (self.tablename,tmpC,tmpA))
                if not retS:
                    raise RuntimeError("cannot add %s to database schema" % tmpC)
        if noAdd:
            return
        # check whole schema just in case
        self.checkSchema(noAdd=True)



    # create table
    def createTable(self):
        # ver 0_1_1
        sql  = "CREATE TABLE %s (" % self.tablename
        sql += "'id'             INTEGER PRIMARY KEY,"
        sql += "'JobID'          INTEGER,"
        sql += "'PandaID'        TEXT,"
        sql += "'jobStatus'      TEXT,"
        sql += "'site'           VARCHAR(128),"
        sql += "'cloud'          VARCHAR(20),"
        sql += "'jobType'        VARCHAR(20),"
        sql += "'jobName'        VARCHAR(128),"
        sql += "'inDS'           TEXT,"
        sql += "'outDS'          TEXT,"
        sql += "'libDS'          VARCHAR(255),"
        sql += "'jobParams'      TEXT,"
        sql += "'retryID'        INTEGER,"
        sql += "'provenanceID'   INTEGER,"
        sql += "'creationTime'   TIMESTAMP,"
        sql += "'lastUpdate'     TIMESTAMP,"
        sql += "'dbStatus'       VARCHAR(20),"
        sql += "'buildStatus'    VARCHAR(20),"
        sql += "'commandToPilot' VARCHAR(20),"
        for tmpC in LocalJobSpec.appended:
            tmpA = LocalJobSpec.appended[tmpC]
            sql += "'%s' %s," % (tmpC,tmpA)
        sql = sql[:-1]
        sql += ")"
        # execute
        retS,retV = self.execute(sql)
        if not retS:
            raise RuntimeError("failed to create %s" % self.tablename)
        # confirm
        if not self.checkTable():
            raise RuntimeError("failed to confirm %s" % self.tablename)


# convert Panda jobs to DB representation
def convertPtoD(pandaJobList,pandaIDstatus,localJob=None,fileInfo={},pandaJobForSiteID=None):
    statusOnly = False
    if localJob is not None:
        # update status only
        ddata = localJob
        statusOnly = True
    else:
        # create new spec
        ddata = LocalJobSpec()
    # sort by PandaID
    pandIDs = list(pandaIDstatus)
    pandIDs.sort()
    pStr = ''
    sStr = ''
    ddata.commandToPilot = ''
    for tmpID in pandIDs:
        # PandaID
        pStr += '%s,' % tmpID
        # status
        sStr += '%s,' % pandaIDstatus[tmpID][0]
        # commandToPilot
        if pandaIDstatus[tmpID][1] == 'tobekilled':
            ddata.commandToPilot = 'tobekilled'
    pStr = pStr[:-1]
    sStr = sStr[:-1]
    # job status
    ddata.jobStatus = sStr
    # PandaID
    ddata.PandaID = pStr
    # get panda Job
    pandaJob = None
    if pandaJobList != []:
        # look for buildJob since it doesn't have the first PandaID when retried
        for pandaJob in pandaJobList:
            if pandaJob.prodSourceLabel == 'panda':
                break
    elif pandaJobForSiteID is not None:
        pandaJob = pandaJobForSiteID
    # extract libDS
    if pandaJob is not None:
        if pandaJob.prodSourceLabel == 'panda':
            # build Jobs
            ddata.buildStatus = pandaJob.jobStatus
            for tmpFile in pandaJob.Files:
                if tmpFile.type == 'output':
                    ddata.libDS = tmpFile.dataset
                    break
        else:
            # noBuild or libDS
            ddata.buildStatus = ''
            for tmpFile in pandaJob.Files:
                if tmpFile.type == 'input' and tmpFile.lfn.endswith('.lib.tgz'):
                    ddata.libDS = tmpFile.dataset
                    break
        # release
        ddata.releaseVar = pandaJob.AtlasRelease
        # cache
        tmpCache = re.sub('^[^-]+-*','',pandaJob.homepackage)
        tmpCache = re.sub('_','-',tmpCache)
        ddata.cacheVar = tmpCache
    # return if update status only
    if statusOnly:
        # build job
        if ddata.buildStatus != '':
            ddata.buildStatus = sStr.split(',')[0]
        # set computingSite mainly for rebrokerage
        if pandaJobForSiteID is not None:
            ddata.site = pandaJobForSiteID.computingSite
            ddata.nRebro = pandaJobForSiteID.specialHandling.split(',').count('rebro') + \
                           pandaJobForSiteID.specialHandling.split(',').count('sretry')
        # return
        return ddata
    # job parameters
    ddata.jobParams = pandaJob.metadata
    # extract datasets
    iDSlist  = []
    oDSlist = []
    if fileInfo != {}:
        if 'inDS' in fileInfo:
            iDSlist = fileInfo['inDS']
        if 'outDS' in fileInfo:
            oDSlist = fileInfo['outDS']
    else:
        for pandaJob in pandaJobList:
            for tmpFile in pandaJob.Files:
                if tmpFile.type == 'input' and not tmpFile.lfn.endswith('.lib.tgz'):
                    if tmpFile.dataset not in iDSlist:
                        iDSlist.append(tmpFile.dataset)
                elif tmpFile.type == 'output' and not tmpFile.lfn.endswith('.lib.tgz'):
                    if tmpFile.dataset not in oDSlist:
                        oDSlist.append(tmpFile.dataset)
    # convert to string
    ddata.inDS = ''
    for iDS in iDSlist:
        ddata.inDS += '%s,' % iDS
    ddata.inDS = ddata.inDS[:-1]
    ddata.outDS = ''
    for oDS in oDSlist:
        ddata.outDS += '%s,' % oDS
    ddata.outDS = ddata.outDS[:-1]
    # job name
    ddata.jobName = pandaJob.jobName
    # creation time
    ddata.creationTime = pandaJob.creationTime
    # job type
    ddata.jobType = pandaJob.prodSeriesLabel
    # site
    ddata.site = pandaJob.computingSite
    # cloud
    ddata.cloud = pandaJob.cloud
    # job ID
    ddata.JobID = pandaJob.jobDefinitionID
    # retry ID
    ddata.retryID = 0
    # provenance ID
    ddata.provenanceID = pandaJob.jobExecutionID
    # groupID
    ddata.groupID = pandaJob.jobsetID
    ddata.retryJobsetID = -1
    if pandaJob.sourceSite not in ['NULL',None,'']:
        ddata.parentJobsetID = long(pandaJob.sourceSite)
    else:
        ddata.parentJobsetID = -1
    # job type
    ddata.jobType = pandaJob.processingType
    # the number of rebrokerage actions
    ddata.nRebro = pandaJob.specialHandling.split(',').count('rebro')
    # jediTaskID
    ddata.jediTaskID = -1
    # return
    return ddata



# convert JediTask to DB representation
def convertJTtoD(jediTaskDict,localJob=None):
    statusOnly = False
    if localJob is not None:
        # update status only
        ddata = localJob
        statusOnly = True
    else:
        # create new spec
        ddata = LocalJobSpec()
    # max IDs
    maxIDs = 20
    # task status
    ddata.taskStatus = jediTaskDict['status']
    # statistic
    ddata.jobStatus = jediTaskDict['statistics']
    # PandaID
    ddata.PandaID = ''
    for tmpPandaID in jediTaskDict['PandaID'][:maxIDs]:
        ddata.PandaID += '%s,' % tmpPandaID
    ddata.PandaID = ddata.PandaID[:-1]
    if len(jediTaskDict['PandaID']) > maxIDs:
        ddata.PandaID += ',+%sIDs' % (len(jediTaskDict['PandaID'])-maxIDs)
    # merge status
    if 'mergeStatus' not in jediTaskDict or jediTaskDict['mergeStatus'] is None:
        ddata.mergeJobStatus = 'NA'
    else:
        ddata.mergeJobStatus = jediTaskDict['mergeStatus']
    # merge PandaID
    ddata.mergeJobID = ''
    for tmpPandaID in jediTaskDict['mergePandaID'][:maxIDs]:
        ddata.mergeJobID += '%s,' % tmpPandaID
    ddata.mergeJobID = ddata.mergeJobID[:-1]
    if len(jediTaskDict['mergePandaID']) > maxIDs:
        ddata.mergeJobID += ',+%sIDs' % (len(jediTaskDict['mergePandaID'])-maxIDs)
    # return if update status only
    if statusOnly:
        return ddata
    # release
    ddata.releaseVar = jediTaskDict['transUses']
    # cache
    if jediTaskDict['transHome'] is None:
        tmpCache = ''
    else:
        tmpCache = re.sub('^[^-]+-*','',jediTaskDict['transHome'])
        tmpCache = re.sub('_','-',tmpCache)
    ddata.cacheVar = tmpCache
    # job parameters
    try:
        if isinstance(jediTaskDict['cliParams'],unicode):
            ddata.jobParams = jediTaskDict['cliParams'].encode('utf_8')
        else:
            ddata.jobParams = jediTaskDict['cliParams']
        # truncate
        ddata.jobParams = ddata.jobParams[:1024]
    except Exception:
        pass
    # input datasets
    try:
        # max number of datasets to show
        maxDS = 20
        inDSs = jediTaskDict['inDS'].split(',')
        strInDS = ''
        # concatenate
        for tmpInDS in inDSs[:maxDS]:
            strInDS += "%s," % tmpInDS
        strInDS = strInDS[:-1]
        # truncate
        if len(inDSs) > maxDS:
            strInDS += ',+{0}DSs'.format(len(inDSs)-maxDS)
        ddata.inDS = strInDS
    except Exception:
        ddata.inDS = jediTaskDict['inDS']
    # output datasets
    ddata.outDS = jediTaskDict['outDS']
    # job name
    ddata.jobName = jediTaskDict['taskName']
    # creation time
    ddata.creationTime = jediTaskDict['creationDate']
    # job type
    ddata.jobType = jediTaskDict['processingType']
    # site
    ddata.site = jediTaskDict['site']
    # cloud
    ddata.cloud = jediTaskDict['cloud']
    # job ID
    ddata.JobID = jediTaskDict['reqID']
    # retry ID
    ddata.retryID = 0
    # provenance ID
    ddata.provenanceID = 0
    # groupID
    ddata.groupID = jediTaskDict['reqID']
    # jediTaskID
    ddata.jediTaskID = jediTaskDict['jediTaskID']
    # IDs for retry
    ddata.retryJobsetID = -1
    ddata.parentJobsetID = -1
    # the number of rebrokerage actions
    ddata.nRebro = 0
    # return
    return ddata



# instantiate database proxy
pdbProxy = PdbProxy()


# just initialize DB
def initialzieDB(verbose=False,restoreDB=False):
    if restoreDB:
        pdbProxy.deleteDatabase()
    pdbProxy.initialize()
    pdbProxy.setVerbose(verbose)


# insert job info to DB
def insertJobDB(job,verbose=False):
    tmpLog = PLogger.getPandaLogger()
    # set update time
    job.lastUpdate = datetime.datetime.utcnow()
    # make sql
    sql1 = "INSERT INTO %s (%s) " % (pdbProxy.tablename,LocalJobSpec.columnNames())
    sql1+= "VALUES " + job.values()
    status,out = pdbProxy.execute_direct(sql1)
    if not status:
        raise RuntimeError("failed to insert job")


# update job info in DB
def updateJobDB(job,verbose=False,updateTime=None):
    # make sql
    sql1  = "UPDATE %s SET " % pdbProxy.tablename
    sql1 += job.values(forUpdate=True)
    sql1 += " WHERE JobID=%s " % job.JobID
    # set update time
    if updateTime is not None:
        job.lastUpdate = updateTime
        sql1 += " AND lastUpdate<'%s' " % updateTime.strftime('%Y-%m-%d %H:%M:%S')
    else:
        job.lastUpdate = datetime.datetime.utcnow()
    status,out = pdbProxy.execute_direct(sql1)
    if not status:
        raise RuntimeError("failed to update job")


# set retryID
def setRetryID(job,verbose=False):
    # make sql
    sql1  = "UPDATE %s SET " % pdbProxy.tablename
    sql1 += "retryID=%s,retryJobsetID=%s " % (job.JobID,job.groupID)
    sql1 += " WHERE JobID=%s AND (nRebro IS NULL OR nRebro=%s)" % (job.provenanceID,job.nRebro)
    status,out = pdbProxy.execute(sql1)
    if not status:
        raise RuntimeError("failed to set retryID")


# delete old jobs
def deleteOldJobs(days,verbose=False):
    # time limit
    limit = datetime.datetime.utcnow() - datetime.timedelta(days=days)
    # make sql
    sql1  = "DELETE FROM %s " % pdbProxy.tablename
    sql1 += " WHERE creationTime<'%s' " % limit.strftime('%Y-%m-%d %H:%M:%S')
    status,out = pdbProxy.execute_direct(sql1)
    if not status:
        raise RuntimeError("failed to delete old jobs")


# read job info from DB
def readJobDB(JobID,verbose=False):
    # make sql
    sql1 = "SELECT %s FROM %s " % (LocalJobSpec.columnNames(),pdbProxy.tablename)
    sql1+= "WHERE JobID=%s" % JobID
    # execute
    status,out = pdbProxy.execute_direct(sql1, fetch=True)
    if not status:
        raise RuntimeError("failed to get JobID=%s" % JobID)
    if len(out) == 0:
        return None
    # instantiate LocalJobSpec
    for values in out:
        job = LocalJobSpec()
        job.pack(values)
        # return frozen job if exists
        if job.dbStatus == 'frozen':
            return job
    # return any
    return job


# read jobset info from DB
def readJobsetDB(JobsetID,verbose=False):
    # make sql
    sql1 = "SELECT %s FROM %s " % (LocalJobSpec.columnNames(),pdbProxy.tablename)
    sql1+= "WHERE groupID=%s" % JobsetID
    # execute
    status,out = pdbProxy.execute(sql1)
    if not status:
        raise RuntimeError("failed to get JobsetID=%s" % JobsetID)
    if len(out) == 0:
        return None
    # instantiate LocalJobSpec
    tmpJobMap = {}
    for tmpStr in out:
        values = tmpStr.split('|')
        job = LocalJobSpec()
        job.pack(values)
        # return frozen job if exists
        if job.dbStatus == 'frozen' or job.JobID not in tmpJobMap:
            tmpJobMap[job.JobID] = job
    # make jobset
    jobset = LocalJobsetSpec()
    # set jobs
    jobset.setJobs(tmpJobMap.values())
    # return any
    return jobset


# check jobset status in DB
def checkJobsetStatus(JobsetID,verbose=False):
    # logger
    tmpLog = PLogger.getPandaLogger()
    # make sql
    sql1 = "SELECT %s FROM %s " % (LocalJobSpec.columnNames(),pdbProxy.tablename)
    sql1+= "WHERE groupID=%s" % JobsetID
    failedRet = False,None
    # execute
    status,out = pdbProxy.execute(sql1)
    if not status:
        tmpLog.error(out)
        tmpLog.error("failed to access local DB")
        return failedRet
    if len(out) == 0:
        tmpLog.error("failed to get JobsetID=%s from local DB" % JobsetID)
        return None
    # instantiate LocalJobSpec
    jobMap = {}
    for tmpStr in out:
        values = tmpStr.split('|')
        job = LocalJobSpec()
        job.pack(values)
        # use frozen job if exists
        if job.JobID not in jobMap or job.dbStatus == 'frozen':
            jobMap[job.JobID] = job
    # check all job status
    for tmpJobID in jobMap:
        tmpJobSpec = jobMap[tmpJobID]
        if tmpJobSpec != 'frozen':
            return True,'running'
    # return
    return True,'frozen'


# bulk read job info from DB
def bulkReadJobDB(verbose=False):
    # make sql
    sql1 = "SELECT %s FROM %s " % (LocalJobSpec.columnNames(),pdbProxy.tablename)
    # execute
    status,out = pdbProxy.execute_direct(sql1, fetch=True)
    if not status:
        raise RuntimeError("failed to get jobs")
    if len(out) == 0:
        return []
    # instantiate LocalJobSpec
    retMap = {}
    jobsetMap = {}
    for values in out:
        job = LocalJobSpec()
        job.pack(values)
        # use frozen job if exists
        if job.JobID not in retMap or job.dbStatus == 'frozen':
            if job.groupID in [0,'0','NULL',-1,'-1']:
                retMap[long(job.JobID)] = job
            else:
                # add jobset
                tmpJobsetID = long(job.groupID)
                if tmpJobsetID not in retMap or tmpJobsetID not in jobsetMap:
                    jobsetMap[tmpJobsetID] = []
                    jobset = LocalJobsetSpec()
                    retMap[tmpJobsetID] = jobset
                # add job
                jobsetMap[tmpJobsetID].append(job)
    # add jobs to jobset
    for tmpJobsetID in jobsetMap:
        tmpJobList = jobsetMap[tmpJobsetID]
        retMap[tmpJobsetID].setJobs(tmpJobList)
    # sort
    ids = list(retMap)
    ids.sort()
    retVal = []
    for id in ids:
        retVal.append(retMap[id])
    # return
    return retVal


# get list of JobID
def getListOfJobIDs(nonFrozen=False,verbose=False):
    # make sql
    sql1 = "SELECT JobID,dbStatus FROM %s " % pdbProxy.tablename
    # execute
    status,out = pdbProxy.execute_direct(sql1, fetch=True)
    if not status:
        raise RuntimeError("failed to get list of JobIDs")
    allList = []
    frozenList = []
    for item in out:
        # extract JobID
        tmpID = long(item[0])
        # status in DB
        tmpStatus = item[-1]
        # keep all jobs
        if tmpID not in allList:
            allList.append(tmpID)
        # keep frozen jobs
        if nonFrozen and tmpStatus == 'frozen':
            if tmpID not in frozenList:
                frozenList.append(tmpID)
    # remove redundant jobs
    retVal = []
    for item in allList:
        if item not in frozenList:
            retVal.append(item)
    # sort
    retVal.sort()
    # return
    return retVal


# get map of jobsetID and JobIDs
def getMapJobsetIDJobIDs(verbose=False):
    # make sql
    sql1 = "SELECT groupID,JobID FROM %s WHERE groupID is not NULL and groupID != 0 and groupID != ''" % pdbProxy.tablename
    # execute
    status,out = pdbProxy.execute(sql1)
    if not status:
        raise RuntimeError("failed to get list of JobIDs")
    allMap = {}
    for item in out:
        # JobsetID
        tmpJobsetID = long(item.split('|')[0])
        # JobID
        tmpJobID = long(item.split('|')[-1])
        # append
        if tmpJobsetID not in allMap:
            allMap[tmpJobsetID] = []
        if tmpJobID not in allMap[tmpJobsetID]:
            allMap[tmpJobsetID].append(tmpJobID)
    # sort
    for tmpKey in allMap.keys():
        allMap[tmpKey].sort()
    # return
    return allMap


# make JobSetSpec
def makeJobsetSpec(jobList):
    jobset = LocalJobsetSpec()
    jobset.setJobs(jobList)
    return jobset


# get map of jobsetID and jediTaskID
def getJobsetTaskMap(verbose=False):
    # make sql
    sql1 = "SELECT groupID,jediTaskID FROM %s WHERE groupID is not NULL and groupID != 0 and groupID != '' and jediTaskID is not null and jediTaskID != ''" % pdbProxy.tablename
    # execute
    status,out = pdbProxy.execute_direct(sql1, fetch=True)
    if not status:
        raise RuntimeError("failed to get list of JobIDs")
    allMap = {}
    for item in out:
        # JobsetID
        tmpJobsetID = long(item[0])
        # JobID
        jediTaskID = long(item[-1])
        # append
        allMap[jediTaskID] = tmpJobsetID
    # return
    return allMap



import os
import re
import sys
import time
try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote
import datetime
import gzip
import copy
import platform

from . import Client
from . import MiscUtils
from . import PLogger

from .MiscUtils import commands_get_status_output, commands_get_output, commands_get_status_output_with_env
try:
    long()
except Exception:
    long = int

# error code
EC_Config    = 10
EC_Post      = 11


cacheProxyStatus = None
cacheVomsStatus = None
cacheActimeStatus = None
cacheVomsFQAN = ''
cacheActime = ''
cacheLastUpdate = None
cacheVomsInfo = None


# reset cache values
def resetCacheValues():
    global cacheProxyStatus
    global cacheVomsStatus
    global cacheActimeStatus
    global cacheVomsFQAN
    global cacheActime
    global cacheLastUpdate
    global cacheVomsInfo
    timeNow = datetime.datetime.utcnow()
    if cacheLastUpdate is None or (timeNow-cacheLastUpdate) > datetime.timedelta(minutes=60):
        cacheLastUpdate = timeNow
        cacheProxyStatus = None
        cacheVomsStatus = None
        cacheActimeStatus = None
        cacheVomsFQAN = ''
        cacheActime = ''
        cacheVomsInfo = None


# get proxy info
def get_proxy_info(force, verbose):
    global cacheVomsInfo
    if force or cacheVomsInfo is None:
        # get logger
        tmpLog = PLogger.getPandaLogger()
        if Client.use_x509_no_grid():
            if 'PANDA_NICKNAME' not in os.environ:
                status = 1
                nickname = ''
                tmpLog.error('PANDA_NICKNAME is not defined')
            else:
                status = 0
                nickname = os.environ['PANDA_NICKNAME']
            cacheVomsInfo = (status, (nickname,))
        elif not Client.use_oidc():
            # check grid-proxy
            gridSrc = Client._getGridSrc()
            com = '%s voms-proxy-info --all --e' % gridSrc
            if verbose:
                tmpLog.debug(com)
            status,out = commands_get_status_output_with_env(com)
            if verbose:
                tmpLog.debug(status % 255)
                tmpLog.debug(out)
            cacheVomsInfo = status,out
        else:
            # OIDC
            uid, groups, nickname = Client.get_user_name_from_token()
            if uid is None:
                status = 1
            else:
                status = 0
            cacheVomsInfo = (status, (uid, groups, nickname))
    return cacheVomsInfo


# new proxy check
def check_proxy(verbose, voms_role, refresh_info=False, generate_new=True):
    status, out = get_proxy_info(refresh_info, verbose)
    if status == 0:
        if voms_role is None:
            return True
        # check role
        for tmpItem in out.split('\n'):
            if not tmpItem.startswith('attribute'):
                continue
            role = voms_role.split(':')[-1]
            if role in tmpItem:
                return True
    if not generate_new or Client.use_oidc() or Client.use_x509_no_grid():
        return False
    # generate proxy
    import getpass
    tmpLog = PLogger.getPandaLogger()
    tmpLog.info("Need to generate a grid proxy")
    gridPassPhrase = getpass.getpass('Enter GRID pass phrase for this identity:\n').replace('$', '\$').replace('"', r'\"')
    gridSrc = Client._getGridSrc()
    com = '%s echo "%s" | voms-proxy-init -pwstdin ' % (gridSrc, gridPassPhrase)
    com_msg = '%s echo "*****" | voms-proxy-init -pwstdin ' % gridSrc
    if voms_role is None:
        com += '-voms atlas'
        com_msg += '-voms atlas'
    else:
        com += '-voms %s' % voms_role
        com_msg += '-voms %s' % voms_role
    if verbose:
        tmpLog.debug(com_msg)
    status, output = commands_get_status_output_with_env(com)
    if status != 0:
        tmpLog.error(output)
        tmpLog.error("Could not generate a grid proxy")
        sys.exit(EC_Config)
    return check_proxy(verbose, voms_role, refresh_info=True, generate_new=False)



# get nickname
def getNickname(verbose=False):
    nickName = ''
    status, output = get_proxy_info(False, verbose)
    # OIDC
    if Client.use_oidc():
        return output[2]
    # x509 without grid
    if Client.use_x509_no_grid():
        return output[0]
    # X509
    for line in output.split('\n'):
        if line.startswith('attribute'):
            match = re.search('nickname =\s*([^\s]+)\s*\(atlas\)',line)
            if match is not None:
                nickName = match.group(1)
                break
    # check
    if nickName == '':
        # get logger
        tmpLog = PLogger.getPandaLogger()
        wMessage =  'Could not get nickname by using voms-proxy-info which gave\n\n'
        wMessage += output
        wMessage += '\nPlease register nickname to ATLAS VO via\n\n'
        wMessage += '   https://lcg-voms2.cern.ch:8443/voms/atlas/vomrs\n'
        wMessage += '      [Member Info] -> [Edit Personal Info]'
        print('')
        tmpLog.warning(wMessage)
        print('')
    return nickName


# set Rucio accounting
def setRucioAccount(account,appid,forceSet):
    if forceSet or 'RUCIO_ACCOUNT' not in os.environ:
        os.environ['RUCIO_ACCOUNT'] = account
    if forceSet or 'RUCIO_APPID' not in os.environ:
        os.environ['RUCIO_APPID'] = appid


# check name of output dataset
def checkOutDsName(outDS,official,nickName='',mergeOutput=False,verbose=False):
    # get logger
    tmpLog = PLogger.getPandaLogger()
    # check NG chars for SE
    for tmpChar in ['%','|',';','>','<','?','\'','"','(',')','$','@','*',':',
                    '=','&','^','#','\\','@','[',']','{','}','`']:
        if tmpChar in outDS:
            errStr = 'invalid character "%s" is used in --outDS' % tmpChar
            tmpLog.error(errStr)
            return False
    # official dataset
    if official:
        status, output = get_proxy_info(False, verbose)
        # extract production role
        prodGroups = []
        for tmpLine in output.split('\n'):
            match = re.search('/([^/]+)/Role=production',tmpLine)
            if match is not None:
                # ignore atlas production role
                if not match.group(1) in ['atlas']:
                    prodGroups.append(match.group(1))
        # no production role
        if prodGroups == []:
            errStr  = "The --official option requires production role. Please use the --voms option to set production role;\n"
            errStr += "  e.g.,  --voms atlas:/atlas/phys-higgs/Role=production\n"
            errStr += "If you don't have production role for the group please request it in ATLAS VO first"
            tmpLog.error(errStr)
            return False
        # loop over all prefixes
        allowedPrefix = ['group']
        for tmpPrefix in allowedPrefix:
            for tmpGroup in prodGroups:
                tmpPattO = '^'+tmpPrefix+'\d{2}'+'\.'+tmpGroup+'\.'
                tmpPattN = '^'+tmpPrefix+'\.'+tmpGroup+'\.'
                if re.search(tmpPattO,outDS) is not None or re.search(tmpPattN,outDS) is not None:
                    return True
        # didn't match
        errStr  = "Your proxy is allowed to produce official datasets\n"
        errStr += "        with the following prefix\n"
        for tmpPrefix in allowedPrefix:
            for tmpGroup in prodGroups:
                tmpPattO = '%s%s.%s' % (tmpPrefix,time.strftime('%y',time.gmtime()),tmpGroup)
                errStr += "          %s\n" % tmpPattO
                tmpPattN = '%s.%s' % (tmpPrefix,tmpGroup)
                errStr += "          %s\n" % tmpPattN
        errStr += "If you have production role for another group please use the --voms option to set the role\n"
        errStr += "  e.g.,  --voms atlas:/atlas/phys-higgs/Role=production\n"
        tmpLog.error(errStr)
        return False
    # check output dataset format
    matStrN = '^user\.'+nickName+'\.'
    if nickName == '' or re.match(matStrN,outDS) is None:
        if nickName == '':
            errStr = "Could not get nickname from voms proxy\n"
        else:
            outDsPrefixN = 'user.%s' % nickName
            errStr  = "outDS must be '%s.<user-controlled string...>'\n" % outDsPrefixN
            errStr += "        e.g., %s.test1234" % outDsPrefixN
        tmpLog.error(errStr)
        return False
    # check length. 200=255-55. 55 is reserved for Panda-internal (_subXYZ etc)
    maxLength = 200
    if mergeOutput:
        maxLengthCont = 120
    else:
        maxLengthCont = 132
    if outDS.endswith('/'):
        # container
        if len(outDS) > maxLengthCont:
            tmpErrStr  = "The name of the output dataset container is too long (%s). " % len(outDS)
            tmpErrStr += "The length must be less than %s " % maxLengthCont
            if mergeOutput:
                tmpErrStr += "when --mergeOutput is used. "
            else:
                tmpErrStr += ". "
            tmpErrStr += "Please note that the limit on the name length is tighter for containers than datasets"
            tmpLog.error(tmpErrStr)
            return False
    else:
        # dataset
        if len(outDS) > maxLength:
            tmpLog.error("output datasetname is too long (%s). The length must be less than %s" % \
                         (len(outDS),maxLength))
            return False
    return True


# convert sys.argv to string
def convSysArgv(argv=None):
    if argv is None:
        argv = sys.argv
    # job params
    if 'PANDA_EXEC_STRING' in os.environ:
        paramStr = os.environ['PANDA_EXEC_STRING']
    else:
        paramStr = argv[0].split('/')[-1]
    for item in argv[1:]:
        # remove option
        match = re.search('(^-[^=]+=)(.+)',item)
        noSpace = False
        if match is not None:
            paramStr += ' %s' % match.group(1)
            item = match.group(2)
            noSpace = True
        if not noSpace:
            paramStr += ' '
        match = re.search('(\*| |\')',item)
        if match is None:
            # normal parameters
            paramStr += '%s' % item
        else:
            # quote string
            paramStr += '"%s"' % item
    # return
    return paramStr


# compare version
def isLatestVersion(latestVer):
    # extract local version numbers
    import PandaToolsPkgInfo
    match = re.search('^(\d+)\.(\d+)\.(\d+)$',PandaToolsPkgInfo.release_version)
    if match is None:
        return True
    localMajorVer  = int(match.group(1))
    localMinorVer  = int(match.group(2))
    localBugfixVer = int(match.group(3))
    # extract local version numbers
    match = re.search('^(\d+)\.(\d+)\.(\d+)$',latestVer)
    if match is None:
        return True
    latestMajorVer  = int(match.group(1))
    latestMinorVer  = int(match.group(2))
    latestBugfixVer = int(match.group(3))
    # compare
    if latestMajorVer > localMajorVer:
        return False
    if latestMajorVer < localMajorVer:
        return True
    if latestMinorVer > localMinorVer:
        return False
    if latestMinorVer < localMinorVer:
        return True
    if latestBugfixVer > localBugfixVer:
        return False
    # latest or higher
    return True


# check panda-client version
def checkPandaClientVer(verbose):
    # get logger
    tmpLog = PLogger.getPandaLogger()
    # get latest version number
    vStatus,latestVer = Client.getPandaClientVer(verbose)
    if vStatus == 0:
        # check version
        if not isLatestVersion(latestVer):
            warStr = "A newer version of panda-client is available at https://twiki.cern.ch/twiki/bin/view/Atlas/PandaTools."
            if os.environ['PANDA_SYS'].startswith('/afs/cern.ch/atlas/offline/external/GRID/DA/panda-client'):
                # if the user uses CERN AFS
                warStr += " Please execute 'source /afs/cern.ch/atlas/offline/external/GRID/DA/panda-client/latest/etc/panda/panda_setup.[c]sh"
            else:
                warStr += " Please execute '%s --update' if you installed the package locally" % sys.argv[0].split('/')[-1]
            print('')
            tmpLog.warning(warStr+'\n')


# function for path completion
def completePathFunc(text, status):
    # remove white spaces
    text = text.strip()
    # convert ~
    useTilde = False
    if text.startswith('~'):
        useTilde = True
        # keep original
        origText = text
        # convert
        text = os.path.expanduser(text)
    # put / to directories
    if (not text.endswith('/')) and os.path.isdir(text):
        text += '/'
    # list dirs/files
    lsStat,output = commands_get_status_output('ls -d %s*' % text)
    results = []
    if lsStat == 0:
        for tmpItem in output.split('\n'):
            # ignore current and parent dirs
            if tmpItem in ['.','..']:
                continue
            # put /
            if os.path.isdir(tmpItem) and not tmpItem.endswith('/'):
                tmpItem += '/'
            # recover ~
            if useTilde:
                tmpItem = re.sub('^%s' % os.path.expanduser(origText),
                                 origText,tmpItem)
            # append
            results.append(tmpItem)
        # sort
        results.sort()
    # return
    return results[status]


# update package
def updatePackage(verbose=False):
    # get logger
    tmpLog = PLogger.getPandaLogger()
    # get the latest version number
    tmpLog.info('start version check')
    status,output = Client.getPandaClientVer(verbose)
    if status != 0:
        tmpLog.error(output)
        tmpLog.error('failed to get the latest version number : %s' % status)
        return False
    # extract version
    latestVer = output
    # check version
    if isLatestVersion(latestVer):
        tmpLog.info('you are already using the latest version')
        return True
    import PandaToolsPkgInfo
    tmpLog.info('update to %s from %s' % (latestVer,PandaToolsPkgInfo.release_version))
    # set readline for auto-complete
    import readline
    readline.parse_and_bind("tab: complete")
    readline.set_completer(completePathFunc)
    readline.parse_and_bind('set show-all-if-ambiguous On')
    # remove +/~ from delimiters
    curDelimter = readline.get_completer_delims()
    curDelimter = re.sub('\+|/|~','',curDelimter)
    readline.set_completer_delims(curDelimter)
    # installation type
    rpmInstall = False
    newPrefix = os.environ['PANDA_SYS']
    print('')
    print("Please specify type of installation")
    print("   PANDA_SYS=%s" % os.environ['PANDA_SYS'])
    print(" 1. Install to $PANDA_SYS")
    print("      all files in $PANDA_SYS will be erased first and new ones will")
    print("      be installed to the same dir")
    print(" 2. Install to a new dir")
    print("      new files will be installed to somewhere else than $PANDA_SYS")
    print(" 3. Patch (not recommended)")
    print("      existing files in $PANDA_SYS will be patched with new ones")
    print(" 4. RPM installation")
    print("      install RPM. sudo is required")
    print('')
    while True:
        str = input("Enter 1-4 : ")
        if str == '1':
            cleanInstall = True
            break
        if str == '2':
            cleanInstall = False
            # set default
            def startupHookPath():
                defPath = os.environ['PANDA_SYS']
                # remove /
                defPath = re.sub('/+$','',defPath)
                # use one dir up
                defPath = re.sub('/[^/]+$','',defPath)
                # add /
                if not defPath.endswith('/'):
                    defPath += '/'
                # set
                readline.insert_text(defPath)
            # set hook
            readline.set_startup_hook(startupHookPath)
            # get location
            while True:
                newPrefix = input("Enter new location (TAB for autocomplete): ")
                if newPrefix != '':
                    break
            # unset hook
            readline.set_startup_hook(None)
            break
        if str == '3':
            cleanInstall = False
            break
        if str == '4':
            rpmInstall = True
            break
    # get tarball
    tmpLog.info('get panda-client-%s' % latestVer)
    if not rpmInstall:
        packageName = 'panda-client-%s.tar.gz' % latestVer
    else:
        packageName = 'panda-client-%s-1.noarch.rpm' % latestVer
    com = 'wget --no-check-certificate --timeout 120 https://atlpan.web.cern.ch/atlpan/panda-client/%s' \
          % packageName
    status = os.system(com)
    status %= 255
    if status != 0:
        tmpLog.error('failed to download tarball : %s' % status)
        # delete tarball just in case
        commands_get_output('rm %s' % packageName)
        return False
    # install
    if not rpmInstall:
        # expand
        status,output = commands_get_status_output('tar xvfz %s' % packageName)
        status %= 255
        if verbose:
            tmpLog.debug(status)
            tmpLog.debug(output)
        if status != 0:
            tmpLog.error('failed to expand tarball : %s' % status)
            # delete dirs just in case
            commands_get_output('rm -rf panda-client-%s' % latestVer)
            return False
        # delete tarball
        commands_get_output('rm %s' % packageName)
        # save current dir
        currentDir = os.path.realpath(os.getcwd())
        # keep old release
        if cleanInstall:
            tmpLog.info('keep old version in %s.back' % os.environ['PANDA_SYS'])
            backUpDir = '%s.back' % os.environ['PANDA_SYS']
            status,output = commands_get_status_output('rm -rf %s; mv %s %s' % \
                                                     (backUpDir,os.environ['PANDA_SYS'],backUpDir))
            if status != 0:
                tmpLog.error(output)
                tmpLog.error('failed to keep old version')
                # delete dirs
                commands_get_output('rm -rf panda-client-%s' % latestVer)
                return False
        # install
        result = True
        os.chdir('panda-client-%s' % latestVer)
        status,output = commands_get_status_output('python setup.py install --prefix=%s' % newPrefix)
        if verbose:
            tmpLog.debug(output)
            tmpLog.debug(status)
        os.chdir(currentDir)
        status %= 255
        if status != 0:
            tmpLog.error('failed to install panda-client : %s' % status)
            # recover old one
            commands_get_output('rm -rf %s' % os.environ['PANDA_SYS'])
            commands_get_output('mv %s.back %s' % (os.environ['PANDA_SYS'],os.environ['PANDA_SYS']))
            result = False
        # cleanup
        commands_get_output('rm -rf panda-client-%s' % latestVer)
    else:
        # rpm install
        result = True
        newPrefix = ''
        com = 'sudo rpm -Uvh %s' % packageName
        print(com)
        status = os.system(com)
        status %= 255
        if status != 0:
            tmpLog.error('failed to install rpm : %s' % status)
            result = False
        # cleanup
        commands_get_output('rm -rf %s' % packageName)
    # return
    if result:
        tmpLog.info('completed')
        tmpLog.info("please do 'source %s/etc/panda/panda_setup.[c]sh'" % newPrefix)
    return result


# check unmerge dataset
def checkUnmergedDataset(inDS,secDS):
    dsList = []
    if inDS != '':
        dsList += inDS.split(',')
    if secDS != '':
        dsList += secDS.split(',')
    # pattern for unmerged datasets
    unPatt = '^mc.+\.recon\.AOD\.'
    unMergedDs = ''
    for tmpDs in dsList:
        # check dataset name
        if re.search(unPatt,tmpDs) is not None:
            unMergedDs += '%s,' % tmpDs
    unMergedDs = unMergedDs[:-1]
    # return
    if unMergedDs != '':
        # get logger
        tmpLog = PLogger.getPandaLogger()
        msg = "%s is unmerged AOD dataset which is not distributed. Please use mc*.merge.AOD.* instead\n" % unMergedDs
        print('')
        tmpLog.warning(msg)
    return


# read dataset names from text
def readDsFromFile(txtName):
    dsList = ''
    try:
        # read lines
        txt = open(txtName)
        for tmpLine in txt:
            # remove \n
            tmpLine = re.sub('\n','',tmpLine)
            # remove white spaces
            tmpLine = tmpLine.strip()
            # skip comment or empty
            if tmpLine.startswith('#') or tmpLine == '':
                continue
            # append
            dsList += '%s,' % tmpLine
        # close file
        txt.close()
        # remove the last comma
        dsList = dsList[:-1]
    except Exception:
        errType,errValue = sys.exc_info()[:2]
        tmpLog = PLogger.getPandaLogger()
        tmpLog.error('cannot read datasets from %s due to %s:%s' \
                     % (txtName,errType,errValue))
        sys.exit(EC_Config)
    return dsList


# convert param string to JEDI params
def convertParamStrToJediParam(encStr,inputMap,outNamePrefix,encode,padding,usePfnList=False,includeIO=True,
                               extra_in_list=None):
    # list of place holders for input
    inList = ['IN','CAVIN','MININ','LOMBIN','HIMBIN','BHIN','BGIN','BGHIN','BGCIN','BGOIN']
    if extra_in_list:
        inList += extra_in_list
    # place holder for seq_number
    seqHolder = 'SEQNUMBER'
    # place holder for output
    outHolder = 'SN'
    # place holders with extension
    digExList = ['SEQNUMBER', 'FIRSTEVENT']
    allExList = digExList + ['DBR']
    # mapping of client and JEDI place holders
    holders = {'SEQNUMBER' : 'RNDM',
               'DBR'       : 'DB',
               'SKIPEVENTS': 'SKIPEVENTS',
               'FIRSTEVENT': None,
               'MAXEVENTS' : None,
               'SEGMENT_NAME': None
               }
    # replace %XYZ with ${XYZ}
    if includeIO:
        for tmpH in inList:
            encStr = re.sub('%'+tmpH+r'\b', '${'+tmpH+'}', encStr)
    # replace %XYZ with ${newXYZ}
    extensionMap = {}
    for newH in holders:
        oldH = holders[newH]
        # JEDI-only place holders
        if oldH is None:
            oldH = newH
        oldH = '%' + oldH
        # with extension
        if newH in allExList:
            if newH in digExList:
                oldH += '(:|=)(\d+)%{0,1}'
            else:
                oldH += '(:|=)([^ \'\"\}]+)'
            # look for extension
            tmpM = re.search(oldH,encStr)
            if tmpM is not None:
                extensionMap[newH] = tmpM.group(2)
            newH = '${' + newH + '}'
        else:
            newH = '${' + newH + '}'
        encStr = re.sub(oldH,newH,encStr)
    # replace %OUT to outDS${SN}
    if includeIO:
        encStr = re.sub('%OUT',outNamePrefix+'.${'+outHolder+'}',encStr)
    # make pattern for split
    patS  = "("
    allKeys = list(holders)
    if includeIO:
        allKeys += inList
        allKeys += [outHolder]
    for tmpH in allKeys:
        patS += '[^=,\"\' \(\{;]*\$\{' + tmpH + '[^\}]*\}[^,\"\' \)\};]*|'
    patS  = patS[:-1]
    patS += ")"
    # split
    tmpItems = re.split(patS,encStr)
    # make parameters
    jobParams = []
    for tmpItem in tmpItems:
        # check if a place holder
        matchP = re.search('\$\{([^:\}]+)',tmpItem)
        if re.search(patS,tmpItem) is not None and matchP is not None:
            tmpHolder = matchP.group(1)
            # set attributes
            if tmpHolder in inList:
                # use constant since it is templated in another option e.g., -i
                tmpDict = {'type':'constant'}
                if encode:
                    tmpDict['value'] = '${' + tmpHolder + '/E}'
                else:
                    tmpDict['value'] = tmpItem
                # set dataset if PFN list is not used or the stream is not primary
                if not usePfnList or tmpHolder not in ['IN']:
                    tmpDict['param_type'] = 'input'
                    tmpDict['dataset'] = inputMap[tmpHolder]
            elif tmpHolder == seqHolder:
                tmpDict = {'type':'template'}
                tmpDict['value'] = tmpItem
                tmpDict['param_type'] = 'pseudo_input'
                tmpDict['dataset'] = 'seq_number'
                if tmpHolder in extensionMap:
                    try:
                        tmpDict['offset'] = long(extensionMap[tmpHolder])
                    except Exception:
                        pass
            elif tmpHolder == outHolder:
                tmpDict = {'type':'template'}
                tmpDict['value'] = tmpItem
                tmpDict['param_type'] = 'output'
                tmpDict['dataset'] = outNamePrefix + tmpItem.split('}')[-1] + '/'
                tmpDict['container'] = tmpDict['dataset']
            else:
                tmpDict = {'type':'template'}
                tmpDict['value'] = tmpItem
                tmpDict['param_type'] = 'number'
                if tmpHolder in extensionMap:
                    try:
                        tmpDict['offset'] = long(extensionMap[tmpHolder])
                    except Exception:
                        pass
        else:
            # constant
            tmpDict = {'type':'constant'}
            if encode:
                tmpDict['value'] = quote(tmpItem)
            else:
                tmpDict['value'] = tmpItem
        # no padding
        if not padding:
            tmpDict['padding'] = False
        # append
        jobParams.append(tmpDict)
    # return
    return jobParams


# split comma-concatenated list items
def splitCommaConcatenatedItems(oldList):
    if isinstance(oldList, str):
        oldList = [oldList]
    newList = []
    for oldItem in oldList:
        temItems = oldItem.split(',')
        for tmpItem in temItems:
            tmpItem = tmpItem.strip()
            # remove empty
            if tmpItem == '':
                continue
            if tmpItem not in newList:
                newList.append(tmpItem)
    return newList


# upload gzipped file
def uploadGzippedFile(origFileName,currentDir,tmpLog,delFilesOnExit,nosubmit,verbose):
    # open original file
    if origFileName.startswith('/'):
        # absolute path
        tmpIn = open(origFileName, 'rb')
    else:
        # relative path
        tmpIn = open('%s/%s' % (currentDir,origFileName), 'rb')
    # use unique name for gzip
    newFileName = 'pre_%s.dat' % MiscUtils.wrappedUuidGen()
    gzipFullPath = '%s/%s.gz' % (currentDir,newFileName)
    delFilesOnExit.append(gzipFullPath)
    # make gzip
    tmpOut = gzip.open(gzipFullPath,'wb')
    tmpOut.writelines(tmpIn)
    tmpOut.close()
    tmpIn.close()
    # upload
    if not nosubmit:
        tmpLog.info("uploading data file for preprocessing")
        status,out = Client.putFile(gzipFullPath,verbose,useCacheSrv=True,reuseSandbox=False)
        if status != 0 or out != 'True':
            # failed
            print(out)
            tmpLog.error("Failed with %s" % status)
            sys.exit(EC_Post)
    # delete
    os.remove(gzipFullPath)
    # return new filename
    return newFileName


# get PFN list
def getListPFN(pfnFile):
    rFile = open(pfnFile)
    inputFileList = []
    for line in rFile:
        line = re.sub('\n','',line)
        line.strip()
        if line != '' and not line.startswith('#'):
            inputFileList.append(line)
    rFile.close()
    inputFileList.sort()
    if len(inputFileList) == 0:
        # get logger
        tmpLog = PLogger.getPandaLogger()
        tmpLog.error("{0} doesn't contain any PFNs".format(pfnFile))
        sys.exit(EC_Config)
    return inputFileList


# check task parameters
def checkTaskParam(taskParamMap,unlimitNumOutputs):
    # check output dataset names
    maxLengthCont = 132
    maxNumOutputs = 10
    nOutputs = 0
    dict_list = taskParamMap['jobParameters']+[taskParamMap['log']] if 'log' in taskParamMap \
        else taskParamMap['jobParameters']
    for tmpDict in dict_list:
        if tmpDict['type'] == 'template' and tmpDict['param_type'] in ['output','log']:
            if tmpDict['param_type'] == 'output':
                nOutputs += 1
            if len(tmpDict['dataset']) > maxLengthCont:
                tmpErrStr  = "The name of an output or log dataset container (%s) is too long (%s). " % (tmpDict['dataset'],len(tmpDict['dataset']))
                tmpErrStr += "The length must be less than %s following DDM definition. " % maxLengthCont
                tmpErrStr += "Please note that one dataset container is creted per output/log type and "
                tmpErrStr += "each name is <outDS>_<extension made from the output filename>/ or <outDS>.log/. "
                # get logger
                tmpLog = PLogger.getPandaLogger()
                tmpLog.error(tmpErrStr)
                return (EC_Config, tmpErrStr)
    if not unlimitNumOutputs and nOutputs > maxNumOutputs:
        errStr  ='Too many output files (=%s) per job. The default limit is %s. ' % (nOutputs,maxNumOutputs)
        errStr += 'You can remove the constraint by using the --unlimitNumOutputs option. '
        errStr += 'But please note that having too many outputs per job causes a severe load on the system. '
        errStr += 'You may be banned if you carelessly use the option'
        tmpLog = PLogger.getPandaLogger()
        tmpLog.error(errStr)
        return(EC_Config, errStr)
    return (0, None)


# replace input and output
def replaceInputOutput(taskParamMap, inDS, outDS, seqNum):
    newTaskParamMap = copy.deepcopy(taskParamMap)
    if inDS != '':
        oldInDS = taskParamMap['dsForIN']
        subInDSbefore = quote('%DATASET_IN')
        subInDSafter = quote(inDS)
        newTaskParamMap['dsForIN'] = inDS
        for tmpDict in newTaskParamMap['jobParameters']:
            if 'dataset' in tmpDict:
                if tmpDict['dataset'] == oldInDS:
                    tmpDict['dataset'] = inDS
            elif tmpDict['type'] == 'constant':
                tmpDict['value'] = re.sub(subInDSbefore, subInDSafter, tmpDict['value'])
    outDS = re.sub('/$', '', outDS)
    oldOutDS = taskParamMap['taskName']
    oldOutDS = re.sub('/$', '', oldOutDS)
    subOutDSbefore = quote('%DATASET_OUT')
    subOutDSafter = quote(outDS)
    subSeqBefore = quote('%BULKSEQNUMBER')
    subSeqAfter = str(seqNum)
    newTaskParamMap['taskName'] = outDS
    newTaskParamMap['log']['dataset'] = re.sub(oldOutDS, outDS, taskParamMap['log']['dataset'])
    newTaskParamMap['log']['container'] = re.sub(oldOutDS, outDS, taskParamMap['log']['container'])
    newTaskParamMap['log']['value'] = re.sub(oldOutDS, outDS, taskParamMap['log']['value'])
    for tmpDict in newTaskParamMap['jobParameters']:
        if 'dataset' in tmpDict:
            if tmpDict['dataset'].startswith(oldOutDS):
                tmpDict['dataset'] = re.sub(oldOutDS, outDS, tmpDict['dataset'])
                tmpDict['container'] = re.sub(oldOutDS, outDS, tmpDict['container'])
                tmpDict['value'] = re.sub(oldOutDS, outDS, tmpDict['value'])
        elif tmpDict['type'] == 'constant':
            tmpDict['value'] = re.sub(subOutDSbefore, subOutDSafter, tmpDict['value'])
            tmpDict['value'] = re.sub(subSeqBefore, subSeqAfter, tmpDict['value'])
            tmpDict['value'] = re.sub(oldOutDS, outDS, tmpDict['value'])
    return newTaskParamMap


# get OS information
def get_os_information():
    return platform.platform()


# extract voms proxy user name
def extract_voms_proxy_username():
    username = None
    status, output = get_proxy_info(False, False)
    if Client.use_oidc():
        return output[0]
    for line in output.split('\n'):
        if line.startswith('subject'):
            subj = line.split(':', 1)[-1].lstrip()
            user_dn = re.sub(r'(/CN=\d+)+$', '', subj.replace('/CN=proxy', ''))
            username = user_dn.split('=')[-1]
            username = re.sub('[ |_]\d+', '', username)
            username = re.sub("[()']", '', username)
            break
    return username


# warning message when PQ is specified
def get_warning_for_pq(site, excluded_site, tmp_log):
    if site not in ['AUTO', None] or excluded_site:
        tmp_log.warning("The grid queue names could change due to consolidation, migration, etc. "
                "Please check with the command listAnalyPQ to use only online/valid queues "
                "when site and/or excludedSite options are specified.")
    return ''



"""
job specification

"""

import re
import datetime

reserveChangedState = False


class JobSpec(object):
    # attributes
    _attributes = ('PandaID', 'jobDefinitionID', 'schedulerID', 'pilotID', 'creationTime', 'creationHost',
                   'modificationTime', 'modificationHost', 'AtlasRelease', 'transformation', 'homepackage',
                   'prodSeriesLabel', 'prodSourceLabel', 'prodUserID', 'assignedPriority', 'currentPriority',
                   'attemptNr', 'maxAttempt', 'jobStatus', 'jobName', 'maxCpuCount', 'maxCpuUnit', 'maxDiskCount',
                   'maxDiskUnit', 'ipConnectivity', 'minRamCount', 'minRamUnit', 'startTime', 'endTime',
                   'cpuConsumptionTime', 'cpuConsumptionUnit', 'commandToPilot', 'transExitCode', 'pilotErrorCode',
                   'pilotErrorDiag', 'exeErrorCode', 'exeErrorDiag', 'supErrorCode', 'supErrorDiag',
                   'ddmErrorCode', 'ddmErrorDiag', 'brokerageErrorCode', 'brokerageErrorDiag',
                   'jobDispatcherErrorCode', 'jobDispatcherErrorDiag', 'taskBufferErrorCode',
                   'taskBufferErrorDiag', 'computingSite', 'computingElement', 'jobParameters',
                   'metadata', 'prodDBlock', 'dispatchDBlock', 'destinationDBlock', 'destinationSE',
                   'nEvents', 'grid', 'cloud', 'cpuConversion', 'sourceSite', 'destinationSite', 'transferType',
                   'taskID', 'cmtConfig', 'stateChangeTime', 'prodDBUpdateTime', 'lockedby', 'relocationFlag',
                   'jobExecutionID', 'VO', 'pilotTiming', 'workingGroup', 'processingType', 'prodUserName',
                   'nInputFiles', 'countryGroup', 'batchID', 'parentID', 'specialHandling', 'jobsetID',
                   'coreCount', 'nInputDataFiles', 'inputFileType', 'inputFileProject', 'inputFileBytes',
                   'nOutputDataFiles', 'outputFileBytes', 'jobMetrics', 'workQueue_ID', 'jediTaskID',
                   'jobSubStatus', 'actualCoreCount', 'reqID', 'maxRSS', 'maxVMEM', 'maxSWAP', 'maxPSS',
                   'avgRSS', 'avgVMEM', 'avgSWAP', 'avgPSS', 'maxWalltime', 'nucleus', 'eventService',
                   'failedAttempt', 'hs06sec', 'gshare', 'hs06', 'totRCHAR', 'totWCHAR', 'totRBYTES',
                   'totWBYTES', 'rateRCHAR', 'rateWCHAR', 'rateRBYTES', 'rateWBYTES', 'resource_type',
                   'diskIO', 'memory_leak', 'memory_leak_x2', 'container_name', 'job_label'
                   )
    # slots
    __slots__ = _attributes+('Files','_changedAttrs')
    # attributes which have 0 by default
    _zeroAttrs = ('assignedPriority', 'currentPriority', 'attemptNr', 'maxAttempt', 'maxCpuCount', 'maxDiskCount',
                  'minRamCount', 'cpuConsumptionTime', 'pilotErrorCode', 'exeErrorCode', 'supErrorCode', 'ddmErrorCode',
                  'brokerageErrorCode', 'jobDispatcherErrorCode', 'taskBufferErrorCode', 'nEvents', 'relocationFlag',
                  'jobExecutionID', 'nOutputDataFiles', 'outputFileBytes')
    # attribute to be suppressed. They are in another table
    _suppAttrs = ('jobParameters','metadata')
    # mapping between sequence and attr
    _seqAttrMap = {'PandaID': 'ATLAS_PANDA.JOBSDEFINED4_PANDAID_SEQ.nextval'}
    # limit length
    _limitLength = {'ddmErrorDiag'           : 500,
                    'taskBufferErrorDiag'    : 300,
                    'jobDispatcherErrorDiag' : 250,
                    'brokerageErrorDiag'     : 250,
                    'pilotErrorDiag'         : 500,
                    'exeErrorDiag'           : 500,
                    'jobSubStatus'           : 80,
                    'supErrorDiag'           : 250,
                    }
    # tag for special handling
    _tagForSH = {'altStgOut'          : 'ao',
                 'allOkEvents'        : 'at',
                 'notDiscardEvents'   : 'de',
                 'decAttOnFailedES'   : 'df',
                 'dynamicNumEvents'   : 'dy',
                 'fakeJobToIgnore'    : 'fake',
                 'homeCloud'          : 'hc',
                 'inFilePosEvtNum'    : 'if',
                 'inputPrestaging'    : 'ip',
                 'lumiBlock'          : 'lb',
                 'mergeAtOs'          : 'mo',
                 'noExecStrCnv'       : 'nc',
                 'putLogToOS'         : 'po',
                 'registerEsFiles'    : 're',
                 'resurrectConsumers' : 'rs',
                 'requestType'        : 'rt',
                 'jobCloning'         : 'sc',
                 'scoutJob'           : 'sj',
                 'usePrefetcher'      : 'up',
                 'useZipToPin'        : 'uz',
                 'writeInputToFile'   : 'wf',
                 }



    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            object.__setattr__(self,attr,None)
        # files list
        object.__setattr__(self,'Files',[])
        # map of changed attributes
        object.__setattr__(self,'_changedAttrs',{})


    # override __getattribute__ for SQL
    def __getattribute__(self,name):
        ret = object.__getattribute__(self,name)
        if ret is None:
            return "NULL"
        return ret


    # override __setattr__ to collecte the changed attributes
    def __setattr__(self,name,value):
        oldVal = getattr(self,name)
        object.__setattr__(self,name,value)
        newVal = getattr(self,name)
        if name == 'jobStatus':
            if oldVal != newVal:
                self.stateChangeTime = datetime.datetime.utcnow()
        # collect changed attributes
        if oldVal != newVal and name not in self._suppAttrs:
            self._changedAttrs[name] = value


    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self,'_changedAttrs',{})


    # add File to files list
    def addFile(self,file):
        # set owner
        file.setOwner(self)
        # append
        self.Files.append(file)


    # pack tuple into JobSpec
    def pack(self,values):
        for i in range(len(self._attributes)):
            attr= self._attributes[i]
            val = values[i]
            object.__setattr__(self,attr,val)


    # return a tuple of values
    def values(self):
        ret = []
        for attr in self._attributes:
            val = getattr(self,attr)
            ret.append(val)
        return tuple(ret)


    # return map of values
    def valuesMap(self,useSeq=False,onlyChanged=False):
        ret = {}
        for attr in self._attributes:
            if useSeq and attr in self._seqAttrMap:
                continue
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self,attr)
            if val == 'NULL':
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            # jobParameters/metadata go to another table
            if attr in self._suppAttrs:
                val = None
            # truncate too long values
            if attr in self._limitLength:
                if val is not None:
                    val = val[:self._limitLength[attr]]
            ret[':%s' % attr] = val
        return ret


    # return state values to be pickled
    def __getstate__(self):
        state = []
        for attr in self._attributes:
            val = getattr(self,attr)
            state.append(val)
        if reserveChangedState:
            state.append(self._changedAttrs)
        # append File info
        state.append(self.Files)
        return state


    # restore state from the unpickled state values
    def __setstate__(self,state):
        for i in range(len(self._attributes)):
            # schema evolution is supported only when adding attributes
            if i+1 < len(state):
                object.__setattr__(self,self._attributes[i],state[i])
            else:
                object.__setattr__(self,self._attributes[i],'NULL')
        object.__setattr__(self,'Files',state[-1])
        if reserveChangedState:
            object.__setattr__(self,'_changedAttrs',state[-2])
        else:
            object.__setattr__(self,'_changedAttrs',{})


    # return column names for INSERT or full SELECT
    def columnNames(cls):
        ret = ""
        for attr in cls._attributes:
            if ret != "":
                ret += ','
            ret += attr
        return ret
    columnNames = classmethod(columnNames)


    # return expression of values for INSERT
    def valuesExpression(cls):
        ret = "VALUES("
        for attr in cls._attributes:
            ret += "%s"
            if attr != cls._attributes[len(cls._attributes)-1]:
                ret += ","
        ret += ")"
        return ret
    valuesExpression = classmethod(valuesExpression)


    # return expression of bind values for INSERT
    def bindValuesExpression(cls,useSeq=False):
        from pandaserver.config import panda_config
        ret = "VALUES("
        for attr in cls._attributes:
            if useSeq and attr in cls._seqAttrMap:
                if panda_config.backend == 'mysql':
                    # mysql
                    ret += "%s," % "NULL"
                else:
                    # oracle
                    ret += "%s," % cls._seqAttrMap[attr]
            else:
                ret += ":%s," % attr
        ret = ret[:-1]
        ret += ")"
        return ret
    bindValuesExpression = classmethod(bindValuesExpression)


    # return an expression for UPDATE
    def updateExpression(cls):
        ret = ""
        for attr in cls._attributes:
            ret = ret + attr + "=%s"
            if attr != cls._attributes[len(cls._attributes)-1]:
                ret += ","
        return ret
    updateExpression = classmethod(updateExpression)


    # return an expression of bind variables for UPDATE
    def bindUpdateExpression(cls):
        ret = ""
        for attr in cls._attributes:
            ret += '%s=:%s,' % (attr,attr)
        ret = ret[:-1]
        ret += ' '
        return ret
    bindUpdateExpression = classmethod(bindUpdateExpression)


    # comparison function for sort
    def compFunc(cls,a,b):
        iPandaID  = list(cls._attributes).index('PandaID')
        iPriority = list(cls._attributes).index('currentPriority')
        if a[iPriority] > b[iPriority]:
            return -1
        elif a[iPriority] < b[iPriority]:
            return 1
        else:
            if a[iPandaID] > b[iPandaID]:
                return 1
            elif a[iPandaID] < b[iPandaID]:
                return -1
            else:
                return 0
    compFunc = classmethod(compFunc)


    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self._attributes:
            if attr in self._changedAttrs:
                ret += '%s=:%s,' % (attr,attr)
        ret  = ret[:-1]
        ret += ' '
        return ret


    # check if goint to merging
    def produceUnMerge(self):
        for tmpFile in self.Files:
            if tmpFile.isUnMergedOutput():
                return True
        return False



    # truncate string attribute
    def truncateStringAttr(cls,attr,val):
        if attr not in cls._limitLength:
            return val
        if val is None:
            return val
        return val[:cls._limitLength[attr]]
    truncateStringAttr = classmethod(truncateStringAttr)



    # set DDM backend
    def setDdmBackEnd(self,backEnd):
        if self.specialHandling in [None,'']:
            self.specialHandling = 'ddm:'+backEnd
        else:
            if 'ddm:' in self.specialHandling:
                self.specialHandling = re.sub('ddm:[,]+','ddm:'+backEnd,
                                              self.specialHandling)
            else:
                self.specialHandling = self.specialHandling+','+ \
                    'ddm:'+backEnd


    # set LB number
    def setLumiBlockNr(self,lumiBlockNr):
        if self.specialHandling in ['',None,'NULL']:
            self.specialHandling = 'lb:{0}'.format(lumiBlockNr)
        else:
            self.specialHandling += ',lb:{0}'.format(lumiBlockNr)



    # get LB number
    def getLumiBlockNr(self):
        if self.specialHandling is not None:
            for tmpItem in self.specialHandling.split(','):
                if tmpItem.startswith('lb:'):
                    return int(tmpItem.split(':')[-1])
        return None



    # get DDM backend
    def getDdmBackEnd(self):
        if self.specialHandling is None:
            return None
        for tmpItem in self.specialHandling.split(','):
            if tmpItem.startswith('ddm:'):
                return tmpItem.split(':')[-1]
        return None



    # set to accept partial finish
    def setToAcceptPartialFinish(self):
        token = 'ap'
        if self.specialHandling in [None,'']:
            self.specialHandling = token
        else:
            if token not in self.specialHandling.split(','):
                self.specialHandling = self.specialHandling+','+token



    # accept partial finish
    def acceptPartialFinish(self):
        token = 'ap'
        if self.specialHandling in [None,'']:
            return False
        else:
            return token in self.specialHandling.split(',')


    # set home cloud
    def setHomeCloud(self,homeCloud):
        if self.specialHandling in ['',None,'NULL']:
            self.specialHandling = 'hc:{0}'.format(homeCloud)
        else:
            self.specialHandling += ',hc:{0}'.format(homeCloud)



    # get cloud
    def getCloud(self):
        if self.specialHandling is not None:
            for tmpItem in self.specialHandling.split(','):
                if tmpItem.startswith('hc:'): # hc: Home Cloud
                    return tmpItem.split(':')[-1]
        return self.cloud


    # check if cancelled or it's flavor
    def isCancelled(self):
        return self.jobStatus in ['cancelled','closed']



    # get file names which were uploaded to alternative locations
    def altStgOutFileList(self):
        try:
            if self.jobMetrics is not None:
                for item in self.jobMetrics.split():
                    if item.startswith('altTransferred='):
                        return item.split('=')[-1].split(',')
        except Exception:
            pass
        return []



    # get mode for alternative stage-out
    def getAltStgOut(self):
        if self.specialHandling is not None:
            for tmpItem in self.specialHandling.split(','):
                if tmpItem.startswith('{0}:'.format(self._tagForSH['altStgOut'])):
                    return tmpItem.split(':')[-1]
        return None



    # set alternative stage-out
    def setAltStgOut(self,mode):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        # remove old value
        newItems = []
        for tmpItem in items:
            if tmpItem.startswith('{0}:'.format(self._tagForSH['altStgOut'])):
                continue
            newItems.append(tmpItem)
        newItems.append('{0}:{1}'.format(self._tagForSH['altStgOut'],mode))
        self.specialHandling = ','.join(newItems)



    # put log files to OS
    def putLogToOS(self):
        if self.specialHandling is not None:
            return self._tagForSH['putLogToOS'] in self.specialHandling.split(',')
        return False



    # set to put log files to OS
    def setToPutLogToOS(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['putLogToOS'] not in items:
            items.append(self._tagForSH['putLogToOS'])
        self.specialHandling = ','.join(items)



    # write input to file
    def writeInputToFile(self):
        if self.specialHandling is not None:
            return self._tagForSH['writeInputToFile'] in self.specialHandling.split(',')
        return False



    # set to write input to file
    def setToWriteInputToFile(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['writeInputToFile'] not in items:
            items.append(self._tagForSH['writeInputToFile'])
        self.specialHandling = ','.join(items)



    # set request type
    def setRequestType(self,reqType):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        newItems = []
        setFlag = False
        for item in items:
            if not item.startswith(self._tagForSH['requestType']):
                newItems.append(item)
        newItems.append('{0}={1}'.format(self._tagForSH['requestType'],reqType))
        self.specialHandling = ','.join(newItems)



    # sort files
    def sortFiles(self):
        try:
            lfnMap = {}
            for tmpFile in self.Files:
                if tmpFile.lfn not in lfnMap:
                    lfnMap[tmpFile.lfn] = []
                lfnMap[tmpFile.lfn].append(tmpFile)
            lfns = list(lfnMap)
            lfns.sort()
            newFiles = []
            for tmpLFN in lfns:
                for tmpFile in lfnMap[tmpLFN]:
                    newFiles.append(tmpFile)
            self.Files = newFiles
        except Exception:
            pass



    # get zip file map
    def getZipFileMap(self):
        zipMap = dict()
        try:
            if self.jobParameters is not None:
                zipStr = re.search('<ZIP_MAP>(.+)</ZIP_MAP>',self.jobParameters)
                if zipStr is not None:
                    for item in zipStr.group(1).split():
                        zipFile,conFiles = item.split(':')
                        conFiles = conFiles.split(',')
                        zipMap[zipFile] = conFiles
        except Exception:
            pass
        return zipMap



    # suppress execute string conversion
    def noExecStrCnv(self):
        if self.specialHandling is not None:
            return self._tagForSH['noExecStrCnv'] in self.specialHandling.split(',')
        return False



    # set to suppress execute string conversion
    def setNoExecStrCnv(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['noExecStrCnv'] not in items:
            items.append(self._tagForSH['noExecStrCnv'])
        self.specialHandling = ','.join(items)



    # in-file positional event number
    def inFilePosEvtNum(self):
        if self.specialHandling is not None:
            return self._tagForSH['inFilePosEvtNum'] in self.specialHandling.split(',')
        return False



    # set to use in-file positional event number
    def setInFilePosEvtNum(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['inFilePosEvtNum'] not in items:
            items.append(self._tagForSH['inFilePosEvtNum'])
        self.specialHandling = ','.join(items)



    # register event service files
    def registerEsFiles(self):
        if self.specialHandling is not None:
            return self._tagForSH['registerEsFiles'] in self.specialHandling.split(',')
        return False



    # set to register event service files
    def setRegisterEsFiles(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['registerEsFiles'] not in items:
            items.append(self._tagForSH['registerEsFiles'])
        self.specialHandling = ','.join(items)



    # set background-able flag
    def setBackgroundableFlag(self):
        self.jobExecutionID = 0
        if self.prodSourceLabel not in ['managed', 'test']:
            return
        try:
            if self.inputFileBytes / self.maxWalltime > 5000:
                return
        except Exception:
            return
        try:
            if self.coreCount <= 1:
                return
        except Exception:
            return
        if self.currentPriority > 250:
            return
        self.jobExecutionID = 1



    # use prefetcher
    def usePrefetcher(self):
        if self.specialHandling is not None:
            return self._tagForSH['usePrefetcher'] in self.specialHandling.split(',')
        return False



    # set to use prefetcher
    def setUsePrefetcher(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['usePrefetcher'] not in items:
            items.append(self._tagForSH['usePrefetcher'])
        self.specialHandling = ','.join(items)



    # use zip to pin
    def useZipToPin(self):
        if self.specialHandling is not None:
            return self._tagForSH['useZipToPin'] in self.specialHandling.split(',')
        return False



    # set to use zip to pin
    def setUseZipToPin(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['useZipToPin'] not in items:
            items.append(self._tagForSH['useZipToPin'])
        self.specialHandling = ','.join(items)



    # not discard events
    def notDiscardEvents(self):
        if self.specialHandling is not None:
            return self._tagForSH['notDiscardEvents'] in self.specialHandling.split(',')
        return False



    # set not to discard events
    def setNotDiscardEvents(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['notDiscardEvents'] not in items:
            items.append(self._tagForSH['notDiscardEvents'])
        self.specialHandling = ','.join(items)



    # all events are done
    def allOkEvents(self):
        if self.specialHandling is not None:
            return self._tagForSH['allOkEvents'] in self.specialHandling.split(',')
        return False



    # set all events are done
    def setAllOkEvents(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['allOkEvents'] not in items:
            items.append(self._tagForSH['allOkEvents'])
        self.specialHandling = ','.join(items)



    # set scout job flag
    def setScoutJobFlag(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['scoutJob'] not in items:
            items.append(self._tagForSH['scoutJob'])
        self.specialHandling = ','.join(items)



    # check if scout job
    def isScoutJob(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        return self._tagForSH['scoutJob'] in items



    # decrement attemptNr of events only when failed
    def decAttOnFailedES(self):
        if self.specialHandling is not None:
            return self._tagForSH['decAttOnFailedES'] in self.specialHandling.split(',')
        return False



    # set to decrement attemptNr of events only when failed
    def setDecAttOnFailedES(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['decAttOnFailedES'] not in items:
            items.append(self._tagForSH['decAttOnFailedES'])
        self.specialHandling = ','.join(items)



    # set fake flag to ignore in monigoring
    def setFakeJobToIgnore(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['fakeJobToIgnore'] not in items:
            items.append(self._tagForSH['fakeJobToIgnore'])
        self.specialHandling = ','.join(items)


    # remove fake flag to ignore in monigoring
    def removeFakeJobToIgnore(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['fakeJobToIgnore'] in items:
            items.remove(self._tagForSH['fakeJobToIgnore'])
        self.specialHandling = ','.join(items)



    # set task attribute
    def set_task_attribute(self, key, value):
        if not isinstance(self.metadata, list):
            self.metadata = [None, None]
        if len(self.metadata) != 3:
            self.metadata.append({})
        self.metadata[2][key] = value



    # get task attribute
    def get_task_attribute(self, key):
        try:
            return self.metadata[2][key]
        except Exception:
            return None

    # set input prestaging
    def setInputPrestaging(self):
        if self.specialHandling is not None:
            items = self.specialHandling.split(',')
        else:
            items = []
        if self._tagForSH['inputPrestaging'] not in items:
            items.append(self._tagForSH['inputPrestaging'])
        self.specialHandling = ','.join(items)

    # use input prestaging
    def useInputPrestaging(self):
        if self.specialHandling is not None:
            return self._tagForSH['inputPrestaging'] in self.specialHandling.split(',')
        return False

    # to a dictionary
    def to_dict(self):
        ret = {}
        for a in self._attributes:
            v = getattr(self, a)
            if isinstance(v, datetime.datetime):
                v = str(v)
            elif v == 'NULL':
                v = None
            ret[a] = v
        return ret



"""
local jobset specification

"""

import re


class LocalJobsetSpec(object):
    # attributes
    _attributes = ('JobsetID','dbStatus','JobMap','PandaID','inDS','outDS',
                   'parentSetID','retrySetID','creationTime','jobStatus',
                   'jediTaskID','taskStatus')
    # slots
    __slots__ = _attributes + ('flag_showSubstatus','flag_longFormat')


    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            setattr(self,attr,None)
        self.flag_showSubstatus = ''
        self.flag_longFormat = False


    # string format
    def __str__(self):
        # set longFormat when showSubstatus
        if self.flag_showSubstatus != '':
            self.flag_longFormat = True
        # get JobID list
        jobIDs = list(self.JobMap)
        jobIDs.sort()
        # initialize
        firstJob = True
        strFormat = "%15s : %s\n"
        strOut1 =  ""
        strOut2 =  ""
        strOutJob = ''
        strJobID  = ''
        for jobID in jobIDs:
            strJobID += '%s,' % jobID
        strJobID = strJobID[:-1]
        # loop over all jobs
        totalBuild = 0
        totalRun   = 0
        totalMerge = 0
        totalJobStatus = {}
        usingMerge = False
        for jobID in jobIDs:
            job = self.JobMap[jobID]
            if firstJob:
                # get common values from the first jobID
                firstJob = False
                # release
                relStr = ''
                if job.releaseVar not in ['','NULL','None',None]:
                    relStr = job.releaseVar
                # cache
                cacheStr = ''
                if job.cacheVar not in ['','NULL','None',None]:
                    cacheStr = job.cacheVar
                # common string representation
                if self.isJEDI():
                    strOut1 += strFormat % ("jediTaskID", self.jediTaskID)
                    strOut1 += strFormat % ("taskStatus", self.taskStatus)
                strOut1 += strFormat % ("JobsetID",     self.JobsetID)
                strOut1 += strFormat % ("type",         job.jobType)
                strOut1 += strFormat % ("release",      relStr)
                strOut1 += strFormat % ("cache",        cacheStr)
                #strOut2 += strFormat % ("JobID"  ,      strJobID)
                #strOut2 += strFormat % ("PandaID",      self.PandaID)
                strOut2 += strFormat % ("inDS",         self.inDS)
                strOut2 += strFormat % ("outDS",        self.outDS)
                if not self.isJEDI():
                    strOut2 += strFormat % ("parentSetID",  self.parentSetID)
                    strOut2 += strFormat % ("retrySetID",   self.retrySetID)
                strOut2 += strFormat % ("creationTime", job.creationTime.strftime('%Y-%m-%d %H:%M:%S'))
                strOut2 += strFormat % ("lastUpdate",   job.lastUpdate.strftime('%Y-%m-%d %H:%M:%S'))
                strOut2 += strFormat % ("params",       job.jobParams)
                if not self.isJEDI():
                    strOut2 += strFormat % ("status",   self.dbStatus)
                else:
                    strOut2 += strFormat % ("inputStatus",'')
            # job status
            statusMap = {}
            for item in job.jobStatus.split(','):
                match = re.search('^(\w+)\*(\d+)$',item)
                if match is None:
                    # non compact
                    if item not in statusMap:
                        statusMap[item] = 0
                    statusMap[item] += 1
                else:
                    # compact
                    tmpStatus = match.group(1)
                    tmpCount  = int(match.group(2))
                    if tmpStatus not in statusMap:
                        statusMap[tmpStatus] = 0
                    statusMap[tmpStatus] += tmpCount
            # merge
            if job.mergeJobStatus not in ['NA']:
                usingMerge = True
            # get PandaIDs for each status
            pandaIDstatusMap = {}
            if not self.isJEDI():
                tmpStatusList  = job.jobStatus.split(',')
                tmpPandaIDList = job.PandaID.split(',')
                for tmpIndex,tmpPandaID in enumerate(tmpPandaIDList):
                    if tmpIndex < len(tmpStatusList):
                        tmpStatus = tmpStatusList[tmpIndex]
                    else:
                        # use unknown for out-range
                        tmpStatus = 'unknown'
                    # append for all jobs
                    if tmpStatus not in totalJobStatus:
                        totalJobStatus[tmpStatus] = 0
                    totalJobStatus[tmpStatus] += 1
                    # status of interest
                    if tmpStatus not in self.flag_showSubstatus.split(','):
                        continue
                    # append for individual job
                    if tmpStatus not in pandaIDstatusMap:
                        pandaIDstatusMap[tmpStatus] = 'PandaID='
                    pandaIDstatusMap[tmpStatus] += '%s,' % tmpPandaID
            else:
                totalJobStatus = statusMap
            statusStr = job.dbStatus
            for tmpStatus in statusMap:
                tmpCount = statusMap[tmpStatus]
                statusStr += '\n%8s   %10s : %s' % ('',tmpStatus,tmpCount)
                if self.flag_showSubstatus:
                    if tmpStatus in pandaIDstatusMap:
                        statusStr += '\n%8s   %10s   %s' % ('','',pandaIDstatusMap[tmpStatus][:-1])
            # number of jobs
            nJobs = len(job.PandaID.split(','))
            if job.buildStatus != '':
                # including buildJob
                nJobsStr = "%d + 1(build)" % (nJobs-1)
                totalBuild += 1
                if usingMerge and job.jobType in ['usermerge']:
                    totalMerge += (nJobs-1)
                else:
                    totalRun += (nJobs-1)
            else:
                nJobsStr = "%d" % nJobs
                if usingMerge and job.jobType in ['usermerge']:
                    totalMerge += nJobs
                else:
                    totalRun += nJobs
            # merging
            if self.isJEDI() and job.mergeJobID != '':
                totalMerge += len(job.mergeJobID.split(','))
            # job specific string representation
            if self.flag_longFormat:
                strOutJob += '\n'
                strOutJob += strFormat % ("JobID",        job.JobID)
                #strOutJob += strFormat % ("nJobs",        nJobsStr)
                strOutJob += strFormat % ("site",         job.site)
                strOutJob += strFormat % ("libDS",        str(job.libDS))
                strOutJob += strFormat % ("retryID",      job.retryID)
                strOutJob += strFormat % ("provenanceID", job.provenanceID)
                strOutJob += strFormat % ("jobStatus",    statusStr)
        # number of jobs
        nJobsStr = "%d" % totalRun
        if usingMerge:
            nJobsStr += " + %s(merge)" % totalMerge
        if totalBuild != 0:
            nJobsStr += " + %s(build)" % totalBuild
        #strOut1 += strFormat % ("nJobs",        nJobsStr)
        strOut = strOut1 + strOut2
        # not long format
        if not self.flag_longFormat:
            for tmpStatus in totalJobStatus:
                tmpCount = totalJobStatus[tmpStatus]
                strOut += '%8s   %10s : %s\n' % ('',tmpStatus,tmpCount)
        else:
            strOut += strOutJob
        # disable showSubstatus and longFormat
        self.flag_showSubstatus = ''
        self.flag_longFormat = False
        # return
        return strOut


    # override __getattribute__
    def __getattribute__(self,name):
        # scan all JobIDs to get dbStatus
        if name == 'dbStatus':
            tmpJobs = object.__getattribute__(self,'JobMap')
            runningFlag = False
            for tmpJobID in tmpJobs:
                tmpJob = tmpJobs[tmpJobID]
                if tmpJob.dbStatus == 'killing':
                    return 'killing'
                if tmpJob.dbStatus == 'running':
                    runningFlag = True
            if runningFlag:
                return 'running'
            return 'frozen'
        ret = object.__getattribute__(self,name)
        return ret


    # set jobs
    def setJobs(self,jobs):
        # append
        retryIDs = []
        parentIDs = []
        for job in jobs:
            # set initial parameters
            if self.JobsetID is None:
                self.JobsetID = job.groupID
                self.JobMap = {}
                self.creationTime = job.creationTime
                self.jediTaskID = job.jediTaskID
                self.taskStatus = job.taskStatus
            self.JobMap[job.JobID] = job
            # get parent/retry
            if job.retryJobsetID not in [0,-1,'0','-1']:
                if job.retryJobsetID not in retryIDs:
                    retryIDs.append(job.retryJobsetID)
            if job.parentJobsetID not in [0,-1,'0','-1']:
                if job.parentJobsetID not in parentIDs:
                    parentIDs.append(job.parentJobsetID)
        # set parent/retry
        retryIDs.sort()
        parentIDs.sort()
        self.retrySetID = ''
        for tmpID in retryIDs:
            self.retrySetID += '%s,' % tmpID
        self.retrySetID = self.retrySetID[:-1]
        self.parentSetID = ''
        for tmpID in parentIDs:
            self.parentSetID += '%s,' % tmpID
        self.parentSetID = self.parentSetID[:-1]
        # aggregate some info
        pStr = ''
        sStatus = ''
        strInDS = ''
        strOutDS = ''
        tmpInDSList = []
        tmpOutDSList = []
        jobIDs = list(self.JobMap)
        jobIDs.sort()
        for jobID in jobIDs:
            job = self.JobMap[jobID]
            # PandaID
            tmpPStr = job.encodeCompact(includeMerge=True)['PandaID']
            if tmpPStr != '':
                pStr += tmpPStr
                pStr += ','
            # inDS and outDS
            try:
                for tmpItem in str(job.inDS).split(','):
                    if tmpItem not in tmpInDSList:
                        tmpInDSList.append(tmpItem)
                        strInDS += '%s,' % tmpItem
            except Exception:
                pass
            try:
                for tmpItem in str(job.outDS).split(','):
                    if tmpItem not in tmpOutDSList:
                        tmpOutDSList.append(tmpItem)
                        strOutDS += '%s,' % tmpItem
            except Exception:
                pass
            # set job status
            sStatus += job.jobStatus + ','
        # set
        self.PandaID   = pStr[:-1]
        self.inDS      = strInDS[:-1]
        self.outDS     = strOutDS[:-1]
        self.jobStatus = sStatus[:-1]


    # check if JEDI
    def isJEDI(self):
        if self.jediTaskID in [-1,'-1','']:
            return False
        return True



import pandas as pd
import plotly.express as px
from jupyter_dash import JupyterDash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import plotly.graph_objects as go
from dash.dependencies import Input, Output

from . import queryPandaMonUtils


def show_task(jeditaskid, verbose=False, mode='inline'):
    # get task
    task = queryPandaMonUtils.query_tasks(23518002, verbose=False)[-1][0]
    # get tasks of the user
    tasks = queryPandaMonUtils.query_tasks(username=task['username'], verbose=False)[-1]
    tids = set([x['jeditaskid'] for x in tasks])
    tids.add(jeditaskid)

    # Build App
    app = JupyterDash(__name__)
    app.layout = html.Div([
        html.Div([
            html.H2("TaskID: "),
            dcc.Dropdown(
                id='dropdown_taskid',
                options=[{'label': i, 'value': i} for i in tids],
                value=jeditaskid
            ),],
            style={'display': 'inline-block', 'width': '20%'}
        ),
        html.Div([
            html.Div([
                html.H2('Task Attributes'),
                dash_table.DataTable(id='00_table',
                                     columns=[{'id': 'attribute', 'name': 'attribute'},
                                              {'id': 'value', 'name': 'value'}],
                                     page_action='none',
                                     style_table={'height': '330px', 'overflowY': 'auto'},
                                     style_cell_conditional=[
                                         {
                                             'if': {'column_id': 'value'},
                                             'textAlign': 'left'
                                         },
                                     ]),],
                style={'display': 'inline-block', 'width': '49%', 'float': 'left', 'padding-top': '30px'}
            ),
            html.Div([
                dcc.Graph(id='01_graph'),],
                style={'display': 'inline-block', 'width': '49%'}
            ),
        ],),
        html.Div([
            html.Div([
                dcc.Graph(id='10_graph')],
                style={'display': 'inline-block', 'width': '49%'}),
            html.Div([
                dcc.Graph(id='11_graph')],
                style={'display': 'inline-block', 'width': '49%'})
        ],),
    ])

    # Run app and display result inline in the notebook
    app.run_server(mode=mode)


    @app.callback(
        Output('00_table', 'data'),
        Output('01_graph', 'figure'),
        Output('10_graph', 'figure'),
        Output('11_graph', 'figure'),
        Input('dropdown_taskid', "value")
    )
    def make_elements(jeditaskid):
        verbose = False
        task = queryPandaMonUtils.query_tasks(jeditaskid, verbose=verbose)[-1][0]
        jobs = queryPandaMonUtils.query_jobs(jeditaskid, drop=False, verbose=verbose)[-1]['jobs']
        jobs = pd.DataFrame(jobs)

        # task data
        task_data = [{'attribute': k, 'value': task[k]} for k in task if isinstance(task[k], (str, type(None)))]

        # figures
        site_fig = px.histogram(jobs, x="computingsite", color="jobstatus")
        ram_fig = px.histogram(jobs, x="maxrss")

        exectime_fig = go.Figure()
        legend_set = set()
        for d in jobs.itertuples(index=False):
            if d.jobstatus == 'finished':
                t_color = 'green'
            elif d.jobstatus == 'failed':
                t_color = 'red'
            else:
                t_color = 'orange'
            if d.jobstatus not in legend_set:
                show_legend = True
                legend_set.add(d.jobstatus)
                exectime_fig.add_trace(
                    go.Scatter(
                        x=[d.creationtime, d.creationtime],
                        y=[d.pandaid, d.pandaid],
                        mode="lines",
                        line=go.scatter.Line(color=t_color),
                        showlegend=True,
                        legendgroup=d.jobstatus,
                        name=d.jobstatus,
                        hoverinfo='skip'
                    )
                )
            exectime_fig.add_trace(
                go.Scatter(
                    x=[d.creationtime, d.endtime],
                    y=[d.pandaid, d.pandaid],
                    mode="lines",
                    line=go.scatter.Line(color=t_color),
                    showlegend=False,
                    legendgroup=d.jobstatus,
                    name="",
                    hovertemplate="PandaID: %{y:d}")
            )
        exectime_fig.update_xaxes(range=[jobs['creationtime'].min(), jobs['endtime'].max()],
                                  title_text='Job Lifetime')
        exectime_fig.update_yaxes(range=[jobs['pandaid'].min() * 0.999, jobs['pandaid'].max() * 1.001],
                                  title_text='PandaID')

        return task_data, site_fig, ram_fig, exectime_fig



"""
extract configuration

"""

import re

_prompt = "ConfigExtractor > "
def _printConfig(str):
    print('%s%s' % (_prompt,str))


def _Service(str):
    tmpSvcNew = None
    tmpSvcOld = None
    # get new service
    try:
        svcMgr = theApp.serviceMgr()
        tmpSvcNew = getattr(svcMgr,str)
    except Exception:
        pass
    # get old service
    try:
        tmpSvcOld = Service(str)
    except Exception:
        pass
    # return old one for 12.0.6
    if tmpSvcOld is not None:
        return tmpSvcOld
    return tmpSvcNew

def _Algorithm(str):
    try:
        return Algorithm(str)
    except Exception:
        return None

    
######################
# input

EventSelector = _Service( "EventSelector" )
if hasattr(EventSelector,'InputCollections') and hasattr(EventSelector.InputCollections,'__len__') \
       and len(EventSelector.InputCollections):
    # POOL
    if hasattr(EventSelector,"CollectionType") and hasattr(EventSelector.CollectionType,'__len__') \
           and len(EventSelector.CollectionType) and EventSelector.CollectionType == "ExplicitROOT":
        # tag collection
        _printConfig('Input=COLL')
        # reference
        try:
            if EventSelector.RefName is not None:
                _printConfig('Input=COLLREF %s' % EventSelector.RefName)
        except Exception:
            pass
        # query
        try:
            if EventSelector.Query is not None:
                _printConfig('Input=COLLQUERY %s' % EventSelector.Query)
        except Exception:
            pass
    else:
        # normal POOL
        _printConfig('Input=POOL')
    # file list
    str = 'InputFiles '
    for file in EventSelector.InputCollections:
        str += '%s ' % file.split('/')[-1]
        _printConfig(str)
else:
    # ByteStream
    noInputFlag = True
    # both _Service and Service need to be checked due to Configurable
    compList = []
    try:
        compList.append(_Service( "ByteStreamInputSvc" ))
    except Exception:
        pass
    try:
        compList.append(Service( "ByteStreamInputSvc" ))
    except Exception:
        pass
    for ByteStreamInputSvc in compList:
        if (hasattr(ByteStreamInputSvc,'FullFileName') and hasattr(ByteStreamInputSvc.FullFileName,'__len__')
            and len(ByteStreamInputSvc.FullFileName)) or \
            (hasattr(ByteStreamInputSvc,'FilePrefix') and hasattr(ByteStreamInputSvc.FilePrefix,'__len__')
             and len(ByteStreamInputSvc.FilePrefix)):
            _printConfig('Input=BS')
            noInputFlag = False
            break
    if noInputFlag:
        try:
            eventSelector = _Service("EventSelector")
            if hasattr(eventSelector, 'Input') and hasattr(eventSelector.Input, '__len__') \
                    and eventSelector.Input:
                _printConfig('Input=BS')
                noInputFlag = False
        except Exception:
            pass
    if noInputFlag:
        _printConfig('No Input')


# back navigation
if hasattr(EventSelector,'BackNavigation') and EventSelector.BackNavigation == True:
    _printConfig('BackNavigation=ON')


# minimum bias
minBiasEventSelector = _Service( "minBiasEventSelector" )
if hasattr(minBiasEventSelector,'InputCollections') and hasattr(minBiasEventSelector.InputCollections,'__len__') \
       and len(minBiasEventSelector.InputCollections):
    _printConfig('Input=MINBIAS')


# cavern
cavernEventSelector = _Service( "cavernEventSelector" )
if hasattr(cavernEventSelector,'InputCollections') and hasattr(cavernEventSelector.InputCollections,'__len__') \
       and len(cavernEventSelector.InputCollections):
    _printConfig('Input=CAVERN')


# beam gas
BeamGasEventSelector = _Service( "BeamGasEventSelector" )
if hasattr(BeamGasEventSelector,'InputCollections') and hasattr(BeamGasEventSelector.InputCollections,'__len__') \
       and len(BeamGasEventSelector.InputCollections):
    _printConfig('Input=BEAMGAS')


# beam halo
BeamHaloEventSelector = _Service( "BeamHaloEventSelector" )
if hasattr(BeamHaloEventSelector,'InputCollections') and hasattr(BeamHaloEventSelector.InputCollections,'__len__') \
       and len(BeamHaloEventSelector.InputCollections):
    _printConfig('Input=BEAMHALO')


# condition files
CondProxyProvider = _Service( "CondProxyProvider" )
if hasattr(CondProxyProvider,'InputCollections') and hasattr(CondProxyProvider.InputCollections,'__len__') \
       and len(CondProxyProvider.InputCollections):
    condStr = ''
    for fName in CondProxyProvider.InputCollections:
        if not fName.startswith('LFN:'):
            condStr += "%s," % fName
    if condStr != '':
        retStr = "CondInput %s" % condStr
        retStr = retStr[:-1]    
        _printConfig(retStr)
    

######################
# configurable

_configs = []
seqList = []
try:
    # get all Configurable names
    from AthenaCommon.AlgSequence import AlgSequence
    tmpKeys = AlgSequence().allConfigurables.keys()
    # get AlgSequences
    seqList = [AlgSequence()]
    try:
        for key in tmpKeys:
            # check if it is available via AlgSequence
            if not hasattr(AlgSequence(),key.split('/')[-1]):
                continue
            # get full name
            tmpConf = getattr(AlgSequence(),key.split('/')[-1])
            if hasattr(tmpConf,'getFullName'):
                tmpFullName = tmpConf.getFullName()
                # append AthSequencer
                if tmpFullName.startswith('AthSequencer/'):
                    seqList.append(tmpConf)
    except Exception:
        pass
    # loop over all sequences
    for tmpAlgSequence in seqList:
        # loop over keys
        for key in tmpKeys:
            # check if it is available via AlgSequence
            if not hasattr(tmpAlgSequence,key.split('/')[-1]):
                continue
            # get fullname
            if key.find('/') == -1:
                if hasattr(tmpAlgSequence,key):
                    tmpAlg = getattr(tmpAlgSequence,key)
                    if hasattr(tmpAlg,'getFullName'):
                        _configs.append(getattr(tmpAlgSequence,key).getFullName())
                    elif hasattr(tmpAlg,'getName') and hasattr(tmpAlg,'getType'):
                        # ServiceHandle
                        _configs.append('%s/%s' % (tmpAlg.getType(),tmpAlg.getName()))
                    else:
                        # use short name if it doesn't have getFullName
                        _configs.append(key)
            else:
                _configs.append(key)
except Exception:
    pass


def _getConfig(key):
    if seqList == []: 
        from AthenaCommon.AlgSequence import AlgSequence
        return getattr(AlgSequence(),key.split('/')[-1])
    else:
        for tmpAlgSequence in seqList:
            if hasattr(tmpAlgSequence,key.split('/')[-1]):
                return getattr(tmpAlgSequence,key.split('/')[-1])            

    

######################
# output

# hist
HistogramPersistencySvc=_Service("HistogramPersistencySvc")
if hasattr(HistogramPersistencySvc,'OutputFile') and hasattr(HistogramPersistencySvc.OutputFile,'__len__') \
       and len(HistogramPersistencySvc.OutputFile):
    _printConfig('Output=HIST')
    _printConfig(' Name: %s' % HistogramPersistencySvc.OutputFile) 
    
# ntuple
NTupleSvc = _Service( "NTupleSvc" )
if hasattr(NTupleSvc,'Output') and hasattr(NTupleSvc.Output,'__len__') and len(NTupleSvc.Output):
    # look for streamname 
    for item in NTupleSvc.Output:
        match = re.search("(\S+)\s+DATAFILE",item)
        if match is not None:
            sName = item.split()[0]
            _printConfig('Output=NTUPLE %s' % sName)
            # extract name
            fmatch = re.search("DATAFILE=(\S+)\s",item)
            if fmatch is not None:
                fName = fmatch.group(1)
                fName = re.sub('[\"\']','',fName)
                fName = fName.split('/')[-1]
                _printConfig(' Name: %s'% fName)

streamOutputFiles = {}
ignoreMetaFiles = []

# RDO
foundStreamRD0 = False
key = "AthenaOutputStream/StreamRDO"
if key in _configs:
    StreamRDO = _getConfig( key )
else:
    StreamRDO = _Algorithm( key.split('/')[-1] )
if hasattr(StreamRDO,'OutputFile') and hasattr(StreamRDO.OutputFile,'__len__') and len(StreamRDO.OutputFile):
    streamOutputFiles[key.split('/')[-1]] = StreamRDO.OutputFile
    _printConfig('Output=RDO %s' % StreamRDO.OutputFile)
    _printConfig(' Name: %s'% StreamRDO.OutputFile)
    foundStreamRD0 = True
    ignoreMetaFiles.append(StreamRDO.OutputFile)
                
# ESD
foundStreamESD = False
key = "AthenaOutputStream/StreamESD"
if key in _configs:
    StreamESD = _getConfig( key )
else:
    StreamESD = _Algorithm( key.split('/')[-1] )
if hasattr(StreamESD,'OutputFile') and hasattr(StreamESD.OutputFile,'__len__') and len(StreamESD.OutputFile):
    streamOutputFiles[key.split('/')[-1]] = StreamESD.OutputFile
    _printConfig('Output=ESD %s' % StreamESD.OutputFile)
    _printConfig(' Name: %s'% StreamESD.OutputFile)
    foundStreamESD = True
    ignoreMetaFiles.append(StreamESD.OutputFile)

# AOD
foundStreamAOD = False
key = "AthenaOutputStream/StreamAOD"
if key in _configs:
    StreamAOD = _getConfig( key )
else:
    StreamAOD = _Algorithm( key.split('/')[-1] )
if hasattr(StreamAOD,'OutputFile') and hasattr(StreamAOD.OutputFile,'__len__') and len(StreamAOD.OutputFile):
    streamOutputFiles[key.split('/')[-1]] = StreamAOD.OutputFile
    _printConfig('Output=AOD %s' % StreamAOD.OutputFile)
    _printConfig(' Name: %s'% StreamAOD.OutputFile)
    foundStreamAOD = True
    ignoreMetaFiles.append(StreamAOD.OutputFile)

# TAG    
keys = ["AthenaOutputStream/StreamTAG","RegistrationStream/StreamTAG"]
foundKey = False
for key in keys:
    if key in _configs:
        StreamTAG = _getConfig( key )
        foundKey = True
        break
if not foundKey:
    StreamTAG = _Algorithm( key.split('/')[-1] )
if hasattr(StreamTAG,'OutputCollection') and hasattr(StreamTAG.OutputCollection,'__len__') and \
       len(StreamTAG.OutputCollection):
    _printConfig('Output=TAG')
    _printConfig(' Name: %s'% StreamTAG.OutputCollection)

# TAGCOM    
keys = ["AthenaOutputStream/StreamTAGCOM","RegistrationStream/StreamTAGCOM"]
foundKey = False
for key in keys:
    if key in _configs:
        StreamTAGX = _getConfig( key )
        foundKey = True
        break
if not foundKey:
    StreamTAGX = _Algorithm( key.split('/')[-1] )
if hasattr(StreamTAGX,'OutputCollection') and hasattr(StreamTAGX.OutputCollection,'__len__') and \
       len(StreamTAGX.OutputCollection):
    _printConfig('Output=TAGX %s %s' % (StreamTAGX.name(),StreamTAGX.OutputCollection))
    _printConfig(' Name: %s'% StreamTAGX.OutputCollection)

# AANT
aantStream = []
appStList = []
for alg in theApp.TopAlg+_configs:
    if alg.startswith("AANTupleStream" ):
        aName = alg.split('/')[-1]
        if alg in _configs:
            AANTupleStream = _getConfig(alg)
        else:
            AANTupleStream = Algorithm(aName)
        if hasattr(AANTupleStream.OutputName,'__len__') and len(AANTupleStream.OutputName):
            fName = AANTupleStream.OutputName
            # look for streamname 
            THistSvc = _Service( "THistSvc" )
            if hasattr(THistSvc.Output,'__len__') and len(THistSvc.Output):
                for item in THistSvc.Output:
                    if re.search(fName,item):
                        sName = item.split()[0]
                        # check stream name
                        if hasattr(AANTupleStream,'StreamName'):
                            if AANTupleStream.StreamName != sName:
                                continue
                        aantStream.append(sName)
                        tmpAantKey = (aName,sName,fName)
                        if tmpAantKey not in appStList:
                            _printConfig('Output=AANT %s %s %s' % (aName,sName,fName))
                            _printConfig(' Name: %s'% fName)
                            appStList.append(tmpAantKey)
                        break

# Stream1
key = "AthenaOutputStream/Stream1"
if key in _configs:
    Stream1 = _getConfig( key )
elif hasattr(theApp._streams,key.split('/')[-1]):
    Stream1 = getattr(theApp._streams,key.split('/')[-1])
else:
    Stream1 = _Algorithm( key.split('/')[-1] )
if hasattr(Stream1,'OutputFile') and hasattr(Stream1.OutputFile,'__len__') and len(Stream1.OutputFile):
    if (hasattr(Stream1,'Enable') and Stream1.Enable) or (not hasattr(Stream1,'Enable')):
        streamOutputFiles[key.split('/')[-1]] = Stream1.OutputFile        
        _printConfig('Output=STREAM1 %s' % Stream1.OutputFile)
        _printConfig(' Name: %s'% Stream1.OutputFile)
        ignoreMetaFiles.append(Stream1.OutputFile)

# Stream2
key = "AthenaOutputStream/Stream2"
if key in _configs:
    Stream2 = _getConfig( key )
elif hasattr(theApp._streams,key.split('/')[-1]):
    Stream2 = getattr(theApp._streams,key.split('/')[-1])
else:
    Stream2 = _Algorithm( key.split('/')[-1] )
if hasattr(Stream2,'OutputFile') and hasattr(Stream2.OutputFile,'__len__') and len(Stream2.OutputFile):
    if (hasattr(Stream2,'Enable') and Stream2.Enable) or (not hasattr(Stream2,'Enable')):    
        streamOutputFiles[key.split('/')[-1]] = Stream2.OutputFile
        _printConfig('Output=STREAM2 %s' % Stream2.OutputFile)
        _printConfig(' Name: %s'% Stream2.OutputFile)        
        ignoreMetaFiles.append(Stream2.OutputFile)
        

# General Stream
strGenFName = ''
strGenStream  = ''
strMetaStream = ''
ignoredStreamList = ['Stream1','Stream2','StreamBS','StreamBSFileOutput']
if foundStreamRD0:
    # for old releases where StreamRDO was an algorithm 
    ignoredStreamList += ['StreamRDO']
if foundStreamESD:
    # for streamESD defined as an algorithm 
    ignoredStreamList += ['StreamESD']
if foundStreamAOD:
    # for streamAOD defined as an algorithm     
    ignoredStreamList += ['StreamAOD']
    

desdStreams = {}
try:
    metaStreams = []
    for genStream in theApp._streams.getAllChildren()+AlgSequence().getAllChildren():
        # check name
        fullName = genStream.getFullName()
        if (fullName.split('/')[0] == 'AthenaOutputStream' or fullName.split('/')[0] == 'Athena::RootNtupleOutputStream') \
                and (not fullName.split('/')[-1] in ignoredStreamList):
            if hasattr(genStream,'OutputFile') and hasattr(genStream.OutputFile,'__len__') and len(genStream.OutputFile):
                if (hasattr(genStream,'Enable') and genStream.Enable) or (not hasattr(genStream,'Enable')):
                    # keep meta data
                    if genStream.OutputFile.startswith("ROOTTREE:") or \
                           (hasattr(genStream,'WriteOnFinalize') and genStream.WriteOnFinalize):
                        metaStreams.append(genStream)
                    elif fullName.split('/')[-1].startswith('StreamDESD'):
                        # ignore StreamDESD to treat it as multiple-streams later
                            continue
                    else:
                        strGenStream += '%s:%s,' % (fullName.split('/')[-1],genStream.OutputFile)
                        streamOutputFiles[fullName.split('/')[-1]] = genStream.OutputFile
                        strGenFName = genStream.OutputFile
                        ignoreMetaFiles.append(genStream.OutputFile)
    # associate meta stream
    for mStream in metaStreams:
        metaOutName = mStream.OutputFile.split(':')[-1]
        assStream = None
        # look for associated stream
        for stName in streamOutputFiles:
            stOut = streamOutputFiles[stName]
            if metaOutName == stOut:
                assStream = stName
                break
        # ignore meta stream since renaming is used instead of changing jobO
        if metaOutName in ignoreMetaFiles:
            continue
        # print meta stream
        if assStream is not None:
            _printConfig('Output=META %s %s' % (mStream.getFullName().split('/')[1],assStream))
            _printConfig(' Name: %s'% metaOutName)
except Exception:
    pass
if strGenStream != '':
    strGenStream = strGenStream[:-1]
    _printConfig('Output=STREAMG %s' % strGenStream)
    _printConfig(' Name: %s'% strGenFName)
if desdStreams != {}:
    for tmpStreamName in desdStreams:
        tmpOutFileName = desdStreams[tmpStreamName]
        _printConfig('Output=DESD %s' % tmpStreamName)
        _printConfig(' Name: %s'% tmpOutFileName)

# THIST
userDataSvcStream = {}
usedTHistStreams = []
THistSvc = _Service( "THistSvc" )
if hasattr(THistSvc.Output,'__len__') and len(THistSvc.Output):
    for item in THistSvc.Output:
        sName = item.split()[0]
        if sName not in aantStream:
            # extract name
            fmatch = re.search("DATAFILE=(\S+)\s",item)
            fName = None
            if fmatch is not None:
                fName = fmatch.group(1)
                fName = re.sub('[\"\']','',fName)
                fName = fName.split('/')[-1]
            # keep output of UserDataSvc
            if sName in ['userdataoutputstream'] or sName.startswith('userdataoutputstream'):
                userDataSvcStream[sName] = fName
                continue
            # skip if defined in StreamG
            if strGenFName != '' and fName == strGenFName:
                continue
            _printConfig('Output=THIST %s' % sName)
            if fmatch is not None:
                _printConfig(' Name: %s'% fName)

# ROOT outputs for interactive Athena
import ROOT
fList = ROOT.gROOT.GetListOfFiles()
for index in range(fList.GetSize()):
    if fList[index].GetOption() == 'CREATE':
        _printConfig('Output=IROOT %s' % fList[index].GetName())
        _printConfig(' Name: %s'% fList[index].GetName())

# BS
ByteStreamCnvSvc = _Service("ByteStreamCnvSvc")
if hasattr(ByteStreamCnvSvc,'ByteStreamOutputSvc') and \
       ByteStreamCnvSvc.ByteStreamOutputSvc=="ByteStreamEventStorageOutputSvc":
    _printConfig('Output=BS')
elif hasattr(ByteStreamCnvSvc,'ByteStreamOutputSvcList') and \
         'ByteStreamEventStorageOutputSvc' in ByteStreamCnvSvc.ByteStreamOutputSvcList:
    _printConfig('Output=BS')    

# selected BS
BSESOutputSvc = _Service("BSESOutputSvc")
if hasattr(BSESOutputSvc,'SimpleFileName'):
    _printConfig('Output=SelBS %s' % BSESOutputSvc.SimpleFileName)
    _printConfig(' Name: %s'% BSESOutputSvc.SimpleFileName)

# MultipleStream
try:
    from OutputStreamAthenaPool.MultipleStreamManager import MSMgr
    for tmpStream in MSMgr.StreamList:
        # avoid duplication
        if not tmpStream.Name in streamOutputFiles.keys():
            # remove prefix
            tmpFileBaseName = tmpStream.Stream.OutputFile.split(':')[-1]
            _printConfig('Output=MS %s %s' % (tmpStream.Name,tmpFileBaseName))
            _printConfig(' Name: %s'% tmpFileBaseName)
except Exception:
    pass

# UserDataSvc
if userDataSvcStream != {}:
    for userStName in userDataSvcStream:
        userFileName = userDataSvcStream[userStName]
        findStream = False
        # look for associated stream
        for stName in streamOutputFiles:
            stOut = streamOutputFiles[stName]
            if userFileName == stOut:
                _printConfig('Output=USERDATA %s' % stName)
                findStream = True
                break
        # use THIST if not found
        if not findStream:
            _printConfig('Output=THIST %s' % userStName)
            _printConfig(' Name: %s'% userFileName)
        
        

######################
# random number

AtRndmGenSvc = _Service( "AtRndmGenSvc" )
if hasattr(AtRndmGenSvc,'Seeds') and hasattr(AtRndmGenSvc.Seeds,'__len__') and len(AtRndmGenSvc.Seeds):
    # random seeds
    for item in AtRndmGenSvc.Seeds:
        _printConfig('RndmStream %s' % item.split()[0])
import types        
if hasattr(AtRndmGenSvc,'ReadFromFile') and isinstance(AtRndmGenSvc.ReadFromFile,types.BooleanType) and AtRndmGenSvc.ReadFromFile:
    # read from file
    rndFileName = "AtRndmGenSvc.out"
    if hasattr(AtRndmGenSvc.FileToRead,'__len__') and len(AtRndmGenSvc.FileToRead):
        rndFileName = AtRndmGenSvc.FileToRead
    _printConfig('RndmGenFile %s' % rndFileName)

# G4 random seed        
try:
    if hasattr(SimFlags,'SeedsG4'):
        _printConfig('G4RandomSeeds')
except Exception:
    pass



"""
file specification

"""

class FileSpec(object):
    # attributes
    _attributes = ('rowID','PandaID','GUID','lfn','type','dataset','status','prodDBlock',
                   'prodDBlockToken','dispatchDBlock','dispatchDBlockToken','destinationDBlock',
                   'destinationDBlockToken','destinationSE','fsize','md5sum','checksum','scope')
    # slots
    __slots__ = _attributes+('_owner',)


    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            setattr(self,attr,None)
        # set owner to synchronize PandaID
        self._owner = None


    # override __getattribute__ for SQL and PandaID
    def __getattribute__(self,name):
        # PandaID
        if name == 'PandaID':
            if self._owner is None:
                return 'NULL'
            return self._owner.PandaID
        # others
        ret = object.__getattribute__(self,name)
        if ret is None:
            return "NULL"
        return ret


    # set owner
    def setOwner(self,owner):
        self._owner = owner
        
    
    # return a tuple of values
    def values(self):
        ret = []
        for attr in self._attributes:
            val = getattr(self,attr)
            ret.append(val)
        return tuple(ret)


    # pack tuple into FileSpec
    def pack(self,values):
        for i in range(len(self._attributes)):
            attr= self._attributes[i]
            val = values[i]
            setattr(self,attr,val)


    # return state values to be pickled
    def __getstate__(self):
        state = []
        for attr in self._attributes:
            val = getattr(self,attr)
            state.append(val)
        # append owner info
        state.append(self._owner)
        return state


    # restore state from the unpickled state values
    def __setstate__(self,state):
        for i in range(len(self._attributes)):
            if i+1 < len(state):
                setattr(self,self._attributes[i],state[i])
            else:
                setattr(self,self._attributes[i],'NULL')                
        self._owner = state[-1]
        
        
    # return column names for INSERT
    def columnNames(cls):
        ret = ""
        for attr in cls._attributes:
            if ret != "":
                ret += ','
            ret += attr
        return ret
    columnNames = classmethod(columnNames)


    # return expression of values for INSERT
    def valuesExpression(cls):
        ret = "VALUES("
        for attr in cls._attributes:
            ret += "%s"
            if attr != cls._attributes[len(cls._attributes)-1]:
                ret += ","
        ret += ")"            
        return ret
    valuesExpression = classmethod(valuesExpression)


    # return an expression for UPDATE
    def updateExpression(cls):
        ret = ""
        for attr in cls._attributes:
            ret = ret + attr + "=%s"
            if attr != cls._attributes[len(cls._attributes)-1]:
                ret += ","
        return ret
    updateExpression = classmethod(updateExpression)


        

                       
